# -*- coding: utf-8 -*-
"""
ninjotiff.py

Created on Mon Apr 15 13:41:55 2013

A big amount of the tiff writer are (PFE) from
https://github.com/davidh-ssec/polar2grid by David Hoese

License:
Copyright (C) 2013 Space Science and Engineering Center (SSEC),
 University of Wisconsin-Madison.
 Lars Ørum Rasmussen, DMI.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

Original scripts and automation included as part of this package are
distributed under the GNU GENERAL PUBLIC LICENSE agreement version 3.
Binary executable files included as part of this software package are
copyrighted and licensed by their respective organizations, and
distributed consistent with their licensing terms.

Edited by Christian Kliche (Ernst Basler + Partner) to replace pylibtiff with
a modified version of tifffile.py (created by Christoph Gohlke)
"""

import calendar
import logging
import os
from copy import deepcopy
from datetime import datetime

import numpy as np

import mpop.imageo.formats.writer_options as write_opts
from mpop.imageo.formats import tifffile

log = logging.getLogger(__name__)

#-------------------------------------------------------------------------
#
# Ninjo tiff tags from DWD
#
#-------------------------------------------------------------------------
# Geotiff tags.
GTF_ModelPixelScale = 33550
GTF_ModelTiepoint = 33922

# Ninjo tiff tags
NINJO_TAGS = {
    "NTD_Magic": 40000,
    "NTD_SatelliteNameID": 40001,
    "NTD_DateID": 40002,
    "NTD_CreationDateID": 40003,
    "NTD_ChannelID": 40004,
    "NTD_HeaderVersion": 40005,
    "NTD_FileName": 40006,
    "NTD_DataType": 40007,
    "NTD_SatelliteNumber": 40008,
    "NTD_ColorDepth": 40009,
    "NTD_DataSource": 40010,
    "NTD_XMinimum": 40011,
    "NTD_XMaximum": 40012,
    "NTD_YMinimum": 40013,
    "NTD_YMaximum": 40014,
    "NTD_Projection": 40015,
    "NTD_MeridianWest": 40016,
    "NTD_MeridianEast": 40017,
    "NTD_EarthRadiusLarge": 40018,
    "NTD_EarthRadiusSmall": 40019,
    "NTD_GeodeticDate": 40020,
    "NTD_ReferenceLatitude1": 40021,
    "NTD_ReferenceLatitude2": 40022,
    "NTD_CentralMeridian": 40023,
    "NTD_PhysicValue": 40024,
    "NTD_PhysicUnit": 40025,
    "NTD_MinGrayValue": 40026,
    "NTD_MaxGrayValue": 40027,
    "NTD_Gradient": 40028,
    "NTD_AxisIntercept": 40029,
    "NTD_ColorTable": 40030,
    "NTD_Description": 40031,
    "NTD_OverflightDirection": 40032,
    "NTD_GeoLatitude": 40033,
    "NTD_GeoLongitude": 40034,
    "NTD_Altitude": 40035,
    "NTD_AOSAsimuth": 40036,
    "NTD_LOSAsimuth": 40037,
    "NTD_MaxElevation": 40038,
    "NTD_OverflightTime": 40039,
    "NTD_IsBlackLineCorrection": 40040,
    "NTD_IsAtmosphereCorrected": 40041,
    "NTD_IsCalibrated": 40042,
    "NTD_IsNormalized": 40043,
    "NTD_OriginalHeader": 40044,
    "NTD_IsValueTableAvailable": 40045,
    "NTD_ValueTableStringField": 40046,
    "NTD_ValueTableFloatField": 40047,
    "NTD_TransparentPixel": 50000
}

NINJO_TAGS_INV = dict((v, k) for k, v in NINJO_TAGS.items())

#
# model_pixel_scale_tag_count ? ...
# Sometimes DWD product defines an array of length 2 (instead of 3 (as in geotiff)).
#
MODEL_PIXEL_SCALE_COUNT = int(os.environ.get(
    "GEOTIFF_MODEL_PIXEL_SCALE_COUNT", 3))


#-------------------------------------------------------------------------
#
# Read Ninjo products config file.
#
#-------------------------------------------------------------------------
def get_product_config(product_name, force_read=False):
    """Read Ninjo configuration entry for a given product name.

    :Parameters:
        product_name : str
            Name of Ninjo product.

    :Arguments:
        force_read : Boolean
            Force re-reading config file.

    **Notes**:
        * It will look for a *ninjotiff_products.cfg* in MPOP's
          configuration directory defined by *PPP_CONFIG_DIR*.
        * As an example, see *ninjotiff_products.cfg.template* in
          MPOP's *etc* directory.
    """
    return ProductConfigs()(product_name, force_read)


class _Singleton(type):

    def __init__(cls, name_, bases_, dict_):
        super(_Singleton, cls).__init__(name_, bases_, dict_)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls.instance


class ProductConfigs(object):
    __metaclass__ = _Singleton

    def __init__(self):
        self.read_config()

    def __call__(self, product_name, force_read=False):
        if force_read:
            self.read_config()
        return self._products[product_name]

    @property
    def product_names(self):
        return sorted(self._products.keys())

    def read_config(self):
        from ConfigParser import ConfigParser

        def _eval(val):
            try:
                return eval(val)
            except:
                return str(val)

        filename = self._find_a_config_file()
        log.info("Reading Ninjo config file: '%s'" % filename)

        cfg = ConfigParser()
        cfg.read(filename)
        products = {}
        for sec in cfg.sections():
            prd = {}
            for key, val in cfg.items(sec):
                prd[key] = _eval(val)
            products[sec] = prd
        self._products = products

    @staticmethod
    def _find_a_config_file():
        name_ = 'ninjotiff_products.cfg'
        home_ = os.path.dirname(os.path.abspath(__file__))
        penv_ = os.environ.get('PPP_CONFIG_DIR', '')
        for fname_ in [os.path.join(x, name_) for x in (home_, penv_)]:
            if os.path.isfile(fname_):
                return fname_
        raise ValueError("Could not find a Ninjo tiff config file")


#-------------------------------------------------------------------------
#
# Write Ninjo Products
#
#-------------------------------------------------------------------------
def _get_physic_value(physic_unit):
    # return Ninjo's physics unit and value.
    if physic_unit.upper() in ('K', 'KELVIN'):
        return 'Kelvin', 'T'
    elif physic_unit.upper() in ('C', 'CELSIUS'):
        return 'Celsius', 'T'
    elif physic_unit == '%':
        return physic_unit, 'Reflectance'
    elif physic_unit.upper() in ('MW M-2 SR-1 (CM-1)-1',):
        return physic_unit, 'Radiance'
    else:
        return physic_unit, 'Unknown'


def _get_projection_name(area_def):
    # return Ninjo's projection name.
    proj_name = area_def.proj_dict['proj']
    if proj_name in ('eqc',):
        return 'PLAT'
    elif proj_name in ('merc',):
        return 'MERC'
    elif proj_name in ('stere',):
        lat_0 = area_def.proj_dict['lat_0']
        if lat_0 < 0:
            return 'SPOL'
        else:
            return 'NPOL'
    return None


def _get_pixel_size(projection_name, area_def):
    if projection_name in ['PLAT', 'MERC']:
        upper_left = area_def.get_lonlat(0, 0)
        lower_right = area_def.get_lonlat(
            area_def.shape[0] - 1, area_def.shape[1] - 1)
        pixel_size = abs(lower_right[0] - upper_left[0]) / (area_def.shape[1] - 1),\
            abs(upper_left[1] - lower_right[1]) / (area_def.shape[0] - 1)
    elif projection_name in ('NPOL', 'SPOL'):
        pixel_size = (np.rad2deg(area_def.pixel_size_x / float(area_def.proj_dict['a'])),
                      np.rad2deg(area_def.pixel_size_y / float(area_def.proj_dict['b'])))
    else:
        raise ValueError("Could determine pixel size from projection name '%s'" %
                         projection_name + " (Unknown)")
    return pixel_size


def _get_satellite_altitude(filename):
    # Guess altitude (probably no big deal if we fail).
    sat_altitudes = {'MSG': 36000.0,
                     'METOP': 817.0,
                     'NOAA': 870.0}

    filename = os.path.basename(filename).upper()
    for nam_, alt_ in sat_altitudes.items():
        if nam_ in filename:
            return alt_
    return None


def _finalize(geo_image, dtype=np.uint8, value_range_measurement_unit=None, data_is_scaled_01=True):
    """Finalize a mpop GeoImage for Ninjo. Specialy take care of phycical scale
    and offset.

    :Parameters:
        geo_image : mpop.imageo.geo_image.GeoImage
            See MPOP's documentation.
        dtype : bits per sample np.uint8 or np.uint16 (default: np.uint8)
        value_range_measurement_unit: list or tuple
            Defining minimum and maximum value range. Data will be clipped into
            that range. Default is no clipping and auto scale.
        data_is_scaled_01: boolean
            If true (default), input data is assumed to be in the [0.0, 1.0] range.

    :Returns:
        image : numpy.array
            Final image.
        scale : float
            Scale for transform pixel value to physical value.
        offset : float
            Offset for transform pixel value to physical value.
        fill_value : int
            Value for masked out pixels.

    **Notes**:
        physic_val = image*scale + offset
        Example values for value_range_measurement_unit are (0, 125) or (40.0, -87.5)
    """
    if geo_image.mode == 'L':
        # PFE: mpop.satout.cfscene
        data = geo_image.channels[0]
        fill_value = np.iinfo(dtype).min
        log.debug("Transparent pixel are forced to be %d" % fill_value)
        log.debug("Before scaling: %.2f, %.2f, %.2f" %
                  (data.min(), data.mean(), data.max()))
        if np.ma.count_masked(data) == data.size:
            # All data is masked
            data = np.ones(data.shape, dtype=dtype) * fill_value
            scale = 1
            offset = 0
        else:
            if value_range_measurement_unit and data_is_scaled_01:
                # No additional scaling of the input data - assume that data is
                # within [0.0, 1.0] and interpret 0.0 as
                # value_range_measurement_unit[0] and 1.0 as
                # value_range_measurement_unit[1]

                # Make room for transparent pixel.
                scale_fill_value = (
                    (np.iinfo(dtype).max) / (np.iinfo(dtype).max + 1.0))
                geo_image = deepcopy(geo_image)
                geo_image.channels[0] *= scale_fill_value

                geo_image.channels[0] += 1 / (np.iinfo(dtype).max + 1.0)

                channels, fill_value = geo_image._finalize(dtype)
                data = channels[0]

                scale = ((value_range_measurement_unit[1] -
                          value_range_measurement_unit[0]) /
                         (np.iinfo(dtype).max))
                # Handle the case where all data has the same value.
                scale = scale or 1
                offset = value_range_measurement_unit[0]

                mask = data.mask

                offset -= scale

                if fill_value is None:
                    fill_value = 0

            else:
                if value_range_measurement_unit:
                    data.clip(value_range_measurement_unit[0],
                              value_range_measurement_unit[1], data)
                    chn_min = value_range_measurement_unit[0]
                    chn_max = value_range_measurement_unit[1]
                    log.debug("Scaling, using value range %.2f - %.2f" %
                              (value_range_measurement_unit[0], value_range_measurement_unit[1]))
                else:
                    chn_max = data.max()
                    chn_min = data.min()
                    log.debug("Doing auto scaling")

                # Make room for transparent pixel.
                scale = ((chn_max - chn_min) /
                         (np.iinfo(dtype).max - 1.0))

                # Handle the case where all data has the same value.
                scale = scale or 1
                offset = chn_min

                # Scale data to dtype, and adjust for transparent pixel forced
                # to be minimum.
                mask = data.mask
                data = 1 + ((data.data - offset) / scale).astype(dtype)
                offset -= scale

            data[mask] = fill_value

            if log.getEffectiveLevel() == logging.DEBUG:
                d__ = np.ma.array(data, mask=(data == fill_value))
                log.debug("After scaling:  %.2f, %.2f, %.2f" % (d__.min(),
                                                                d__.mean(),
                                                                d__.max()))
                d__ = data * scale + offset
                d__ = np.ma.array(d__, mask=(data == fill_value))
                log.debug("Rescaling:      %.2f, %.2f, %.2f" % (d__.min(),
                                                                d__.mean(),
                                                                d__.max()))
                del d__

        return data, scale, offset, fill_value

    elif geo_image.mode == 'RGB':
        channels, fill_value = geo_image._finalize(dtype)
        if fill_value is None:
            mask = (np.ma.getmaskarray(channels[0]) &
                    np.ma.getmaskarray(channels[1]) &
                    np.ma.getmaskarray(channels[2]))
            channels.append((np.ma.logical_not(mask) *
                             np.iinfo(channels[0].dtype).max).astype(channels[0].dtype))
            fill_value = (0, 0, 0, 0)

        data = np.dstack([channel.filled(fill_v)
                          for channel, fill_v in zip(channels, fill_value)])
        return data, 1.0, 0.0, fill_value[0]

    elif geo_image.mode == 'RGBA':
        channels, fill_value = geo_image._finalize(dtype)
        fill_value = fill_value or (0, 0, 0, 0)
        data = np.dstack((channels[0].filled(fill_value[0]),
                          channels[1].filled(fill_value[1]),
                          channels[2].filled(fill_value[2]),
                          channels[3].filled(fill_value[3])))
        return data, 1.0, 0.0, fill_value[0]

    elif geo_image.mode == 'P':
        fill_value = 0
        data = geo_image.channels[0]
        if isinstance(data, np.ma.core.MaskedArray):
            data = data.filled(fill_value)
        data = data.astype(dtype)
        log.debug("Value range: %.2f, %.2f, %.2f" %
                  (data.min(), data.mean(), data.max()))
        return data, 1.0, 0.0, fill_value

    else:
        raise ValueError("Don't known how til handle image mode '%s'" %
                         str(geo_image.mode))


def save(geo_image, filename, ninjo_product_name=None, writer_options=None,
         **kwargs):
    """MPOP's interface to Ninjo TIFF writer.

    :Parameters:
        geo_image : mpop.imageo.geo_image.GeoImage
            See MPOP's documentation.
        filename : str
            The name of the TIFF file to be created
    :Keywords:
        ninjo_product_name : str
            Optional index to Ninjo configuration file.
        writer_options : dict
            options dictionary as defined in MPOP interface
            See _write
        kwargs : dict
            See _write

    **Note**:
        * Some arguments are type casted, since they could come from a config file, read as strings.
        * 8 bits grayscale with a colormap (if specified, inverted for IR channels).
        * 16 bits grayscale with no colormap (if specified, MinIsWhite is set for IR).
        * min value will be reserved for transparent color.
        * If possible mpop.imageo.image's standard finalize will be used.
    """

    if writer_options:
        # add writer_options
        kwargs.update(writer_options)
        if 'ninjo_product_name' in writer_options:
            ninjo_product_name = writer_options['ninjo_product_name']

    dtype = np.uint8
    if 'nbits' in kwargs:
        nbits = int(kwargs['nbits'])
        if nbits == 16:
            dtype = np.uint16

    try:
        value_range_measurement_unit = (float(kwargs["ch_min_measurement_unit"]),
                                        float(kwargs["ch_max_measurement_unit"]))
    except KeyError:
        value_range_measurement_unit = None

    data_is_scaled_01 = bool(kwargs.get("data_is_scaled_01", True))

    data, scale, offset, fill_value = _finalize(geo_image,
                                                dtype=dtype,
                                                data_is_scaled_01=data_is_scaled_01,
                                                value_range_measurement_unit=value_range_measurement_unit,)

    area_def = geo_image.area
    time_slot = geo_image.time_slot

    # Some Ninjo tiff names
    kwargs['gradient'] = scale
    kwargs['axis_intercept'] = offset
    kwargs['transparent_pix'] = fill_value
    kwargs['image_dt'] = time_slot
    kwargs['is_calibrated'] = True
    if geo_image.mode == 'P' and 'cmap' not in kwargs:
        r, g, b = zip(*geo_image.palette)
        r = list((np.array(r) * 255).astype(np.uint8))
        g = list((np.array(g) * 255).astype(np.uint8))
        b = list((np.array(b) * 255).astype(np.uint8))
        if len(r) < 256:
            r += [0] * (256 - len(r))
            g += [0] * (256 - len(g))
            b += [0] * (256 - len(b))
        kwargs['cmap'] = r, g, b

    write(data, filename, area_def, ninjo_product_name, **kwargs)


def write(image_data, output_fn, area_def, product_name=None, **kwargs):
    """Generic Ninjo TIFF writer.

    If 'prodcut_name' is given, it will load corresponding Ninjo tiff metadata
    from '${PPP_CONFIG_DIR}/ninjotiff.cfg'. Else, all Ninjo tiff metadata should
    be passed by '**kwargs'. A mixture is allowed, where passed arguments
    overwrite config file.

    :Parameters:
        image_data : 2D numpy array
            Satellite image data to be put into the NinJo compatible tiff
        output_fn : str
            The name of the TIFF file to be created
        area_def: pyresample.geometry.AreaDefinition
            Defintion of area
        product_name : str
            Optional index to Ninjo configuration file.

    :Keywords:
        kwargs : dict
            See _write
    """
    upper_left = area_def.get_lonlat(0, 0)
    lower_right = area_def.get_lonlat(
        area_def.shape[0] - 1, area_def.shape[1] - 1)

    if len(image_data.shape) == 3:
        if image_data.shape[2] == 4:
            shape = (area_def.y_size, area_def.x_size, 4)
            log.info("Will generate RGBA product")
        else:
            shape = (area_def.y_size, area_def.x_size, 3)
            log.info("Will generate RGB product")
        write_rgb = True
    else:
        shape = (area_def.y_size, area_def.x_size)
        write_rgb = False
        log.info("Will generate single band product")

    if image_data.shape != shape:
        raise ValueError("Raster shape %s does not correspond to expected shape %s" % (
            str(image_data.shape), str(shape)))

    # Ninjo's physical units and value.
    # If just a physical unit (e.g. 'C') is passed, it will then be
    # translated into Ninjo's unit and value (e.q 'CELCIUS' and 'T').
    physic_unit = kwargs.get('physic_unit', None)
    if physic_unit and not kwargs.get('physic_value', None):
        kwargs['physic_unit'], kwargs['physic_value'] = \
            _get_physic_value(physic_unit)

    # Ninjo's projection name.
    kwargs['projection'] = kwargs.pop('projection', None) or \
        _get_projection_name(area_def) or \
        area_def.proj_id.split('_')[-1]

    # Get pixel size
    if 'pixel_xres' not in kwargs or 'pixel_yres' not in kwargs:
        kwargs['pixel_xres'], kwargs['pixel_yres'] = \
            _get_pixel_size(kwargs['projection'], area_def)

    # Get altitude.
    altitude = kwargs.pop('altitude', None) or \
        _get_satellite_altitude(output_fn)
    if altitude is not None:
        kwargs['altitude'] = altitude

    if product_name:
        options = deepcopy(get_product_config(product_name))
    else:
        options = {}

    options['meridian_west'] = upper_left[0]
    options['meridian_east'] = lower_right[0]
    if kwargs['projection'].endswith("POL"):
        if 'lat_ts' in area_def.proj_dict:
            options['ref_lat1'] = area_def.proj_dict['lat_ts']
            options['ref_lat2'] = 0
    else:
        if 'lat_0' in area_def.proj_dict:
            options['ref_lat1'] = area_def.proj_dict['lat_0']
            options['ref_lat2'] = 0
    if 'lon_0' in area_def.proj_dict:
        options['central_meridian'] = area_def.proj_dict['lon_0']
    if 'a' in area_def.proj_dict:
        options['radius_a'] = area_def.proj_dict['a']
    if 'b' in area_def.proj_dict:
        options['radius_b'] = area_def.proj_dict['b']
    options['origin_lon'] = upper_left[0]
    options['origin_lat'] = upper_left[1]
    options['min_gray_val'] = image_data.min()
    options['max_gray_val'] = image_data.max()
    options.update(kwargs)  # Update/overwrite with passed arguments

    _write(image_data, output_fn, write_rgb=write_rgb, **options)


# -----------------------------------------------------------------------------
#
# Write tiff file.
#
# -----------------------------------------------------------------------------
def _write(image_data, output_fn, write_rgb=False, **kwargs):
    """Proudly Found Elsewhere (PFE) https://github.com/davidh-ssec/polar2grid
    by David Hoese.

    Create a NinJo compatible TIFF file with the tags used
    by the DWD's version of NinJo.  Also stores the image as tiles on disk
    and creates a multi-resolution/pyramid/overview set of images
    (deresolution: 2,4,8,16).

    :Parameters:
        image_data : 2D or 3D numpy array
            Satellite image data to be put into the NinJo compatible tiff
            An 3D array (HxWx3) is expected for a RGB image.
        filename : str
            The name of the TIFF file to be created

    :Keywords:
        cmap : tuple/list of 3 lists of uint16's
            Individual RGB arrays describing the color value for the
            corresponding data value.  For example, image data with a data
            type of unsigned 8-bit integers have 256 possible values (0-255).
            So each list in cmap will have 256 values ranging from 0 to
            65535 (2**16 - 1). (default linear B&W colormap)
        sat_id : int
            DWD NinJo Satellite ID number
        chan_id : int
            DWD NinJo Satellite Channel ID number
        data_source : str
            String describing where the data came from (SSEC, EUMCAST)
        tile_width : int
            Width of tiles on disk (default 512)
        tile_length : int
            Length of tiles on disk (default 512)
        data_cat : str
            NinJo specific data category
                - data_cat[0] = P (polar) or G (geostat)
                - data_cat[1] = O (original) or P (product)
                - data_cat[2:4] = RN or RB or RA or RN or AN
                                  (Raster, Bufr, ASCII, NIL)

            Example: 'PORN' or 'GORN' or 'GPRN' or 'PPRN'
        pixel_xres : float
            Nadir view pixel resolution in degrees longitude
        pixel_yres : float
            Nadir view pixel resolution in degrees latitude
        origin_lat : float
            Top left corner latitude
        origin_lon : float
            Top left corner longitude
        image_dt : datetime object
            Python datetime object describing the date and time of the image
            data provided in UTC
        projection : str
            NinJo compatible projection name (NPOL,PLAT,etc.)
        meridian_west : float
            Western image border (default 0.0)
        meridian_east : float
            Eastern image border (default 0.0)
        radius_a : float
            Large/equatorial radius of the earth (default <not written>)
        radius_b : float
            Small/polar radius of the earth (default <not written>)
        ref_lat1 : float
            Reference latitude 1 (default <not written>)
        ref_lat2 : float
            Reference latitude 2 (default <not written>)
        central_meridian : float
            Central Meridian (default <not written>)
        physic_value : str
            Physical value type. Examples:
                - Temperature = 'T'
                - Albedo = 'ALBEDO'
        physic_unit : str
            Physical value units. Examples:
                - 'CELSIUS'
                - '%'
        min_gray_val : int
            Minimum gray value (default 0)
        max_gray_val : int
            Maximum gray value (default 255)
        gradient : float
            Gradient/Slope
        axis_intercept : float
            Axis Intercept
        altitude : float
            Altitude of the data provided (default 0.0)
        is_atmo_corrected : bool
            Is the data atmosphere corrected? (True/1 for yes) (default False/0)
        is_calibrated : bool
            Is the data calibrated? (True/1 for yes) (default False/0)
        is_normalized : bool
            Is the data normalized (True/1 for yes) (default False/0)
        description : str
            Description string to be placed in the output TIFF (optional)
        transparent_pix : int
            Transparent pixel value (default -1)
        compression : int
            zlib compression level (default 6)
        inv_def_temperature_cmap : bool (default True)
            Invert the default colormap if physical value type is 'T'
        omit_filename_path : bool (default False)
            Do not store path in NTD_FileName tag

    :Raises:
        KeyError :
            if required keyword is not provided
    """
    def _raise_value_error(text):
        log.error(text)
        raise ValueError(text)

    def _default_colormap(reverse=False, nbits16=False):
        # Basic B&W colormap
        if nbits16:
            if reverse:
                return [[x for x in range(65535, -1, -1)]] * 3
            return [[x for x in range(65536)]] * 3
        else:
            if reverse:
                return [[x * 256 for x in range(255, -1, -1)]] * 3
            return [[x * 256 for x in range(256)]] * 3

    def _eval_or_none(key, eval_func):
        try:
            return eval_func(kwargs[key])
        except KeyError:
            return None

    def _eval_or_default(key, eval_func, default):
        try:
            return eval_func(kwargs[key])
        except KeyError:
            return default

    log.info("Creating output file '%s'" % (output_fn,))

    # Extract keyword arguments
    cmap = kwargs.pop("cmap", None)
    sat_id = int(kwargs.pop("sat_id"))
    chan_id = int(kwargs.pop("chan_id"))
    data_source = str(kwargs.pop("data_source"))
    tile_width = int(kwargs.pop("tile_width", 512))
    tile_length = int(kwargs.pop("tile_length", 512))
    data_cat = str(kwargs.pop("data_cat"))
    pixel_xres = float(kwargs.pop("pixel_xres"))
    pixel_yres = float(kwargs.pop("pixel_yres"))
    origin_lat = float(kwargs.pop("origin_lat"))
    origin_lon = float(kwargs.pop("origin_lon"))
    image_dt = kwargs.pop("image_dt")
    zero_seconds = kwargs.pop("zero_seconds", False)
    projection = str(kwargs.pop("projection"))
    meridian_west = float(kwargs.pop("meridian_west", 0.0))
    meridian_east = float(kwargs.pop("meridian_east", 0.0))
    radius_a = _eval_or_none("radius_a", float)
    radius_b = _eval_or_none("radius_b", float)
    ref_lat1 = _eval_or_none("ref_lat1", float)
    ref_lat2 = _eval_or_none("ref_lat2", float)
    central_meridian = _eval_or_none("central_meridian", float)
    min_gray_val = int(kwargs.pop("min_gray_val", 0))
    max_gray_val = int(kwargs.pop("max_gray_val", 255))
    altitude = _eval_or_none("altitude", float)
    is_blac_corrected = int(bool(kwargs.pop("is_blac_corrected", 0)))
    is_atmo_corrected = int(bool(kwargs.pop("is_atmo_corrected", 0)))
    is_calibrated = int(bool(kwargs.pop("is_calibrated", 0)))
    is_normalized = int(bool(kwargs.pop("is_normalized", 0)))
    inv_def_temperature_cmap = bool(kwargs.pop("inv_def_temperature_cmap", 1))
    omit_filename_path = bool(kwargs.pop("omit_filename_path", 0))
    description = _eval_or_none("description", str)

    physic_value = str(kwargs.pop("physic_value", 'None'))
    physic_unit = str(kwargs.pop("physic_unit", 'None'))
    gradient = float(kwargs.pop("gradient", 1.0))
    axis_intercept = float(kwargs.pop("axis_intercept", 0.0))
    try:
        transparent_pix = int(kwargs.get("transparent_pix", -1))
    except Exception:
        transparent_pix = kwargs.get("transparent_pix")[0]
    finally:
        kwargs.pop("transparent_pix")

    # Keyword checks / verification

    # Handle colormap or not.
    min_is_white = False
    if not write_rgb and not cmap:
        if physic_value == 'T' and inv_def_temperature_cmap:
            reverse = True
        else:
            reverse = False
        if np.iinfo(image_data.dtype).bits == 8:
            # Always generate colormap for 8 bit gray scale.
            cmap = _default_colormap(reverse)
        elif reverse:
            # No colormap for 16 bit gray scale, but for IR, specify white is
            # minimum.
            min_is_white = True

    if cmap and len(cmap) != 3:
        _raise_value_error(
            "Colormap (cmap) must be a list of 3 lists (RGB), not %d" %
            len(cmap))

    if len(data_cat) != 4:
        _raise_value_error("NinJo data type must be 4 characters")
    if data_cat[0] not in ["P", "G"]:
        _raise_value_error(
            "NinJo data type's first character must be 'P' or 'G' not '%s'" %
            data_cat[0])
    if data_cat[1] not in ["O", "P"]:
        _raise_value_error(
            "NinJo data type's second character must be 'O' or 'P' not '%s'" %
            data_cat[1])
    if data_cat[2:4] not in ["RN", "RB", "RA", "BN", "AN"]:
        _raise_value_error(
            "NinJo data type's last 2 characters must be one of %s not '%s'" %
            ("['RN','RB','RA','BN','AN']", data_cat[2:4]))

    if description is not None and len(description) >= 1000:
        log.error("NinJo description must be less than 1000 characters")
        raise ValueError("NinJo description must be less than 1000 characters")

    file_dt = datetime.utcnow()
    file_epoch = calendar.timegm(file_dt.timetuple())
    if zero_seconds:
        log.debug("Applying zero seconds correction")
        image_dt_corr = datetime(image_dt.year, image_dt.month, image_dt.day,
                                 image_dt.hour, image_dt.minute)
    else:
        image_dt_corr = image_dt
    image_epoch = calendar.timegm(image_dt_corr.timetuple())

    compression = _eval_or_default("compression", int, 6)

    def _create_args(image_data, pixel_xres, pixel_yres):
        log.info("creating tags and data for a resolution %dx%d"
                 % image_data.shape[:2])
        args = dict()
        extra_tags = []
        args['extratags'] = extra_tags
        args['software'] = 'tifffile/pytroll'
        args['compress'] = compression

        args['extrasamples_type'] = 2

        # Built ins
        if write_rgb:
            args["photometric"] = 'rgb'
        else:
            if cmap:
                args["photometric"] = 'palette'
                args["colormap"] = [item for sublist in cmap for item in sublist]
            elif min_is_white:
                args["photometric"] = 'miniswhite'
            else:
                args["photometric"] = 'minisblack'

        # planarconfig, samples_per_pixel, orientation, sample_format set by
        # tifffile.py

        args["tile_width"] = tile_width
        args["tile_length"] = tile_length

        # not necessary to set value SMinSampleValue and SMaxsampleValue
        # TIFF6 Spec: The default for SMinSampleValue and SMaxSampleValue is
        # the full range of the data type
        # args["SMinSampleValue"] = 0
        # args["SMaxsampleValue"] = 255

        # NinJo specific tags
        if description is not None:
            extra_tags.append(
                (NINJO_TAGS["NTD_Description"], 's', 0, description, True))

        # Geo tiff tags
        if MODEL_PIXEL_SCALE_COUNT == 3:
            extra_tags.append((GTF_ModelPixelScale,
                               'd', 3, [pixel_xres, pixel_yres, 0.0], True))
        else:
            extra_tags.append((GTF_ModelPixelScale,
                               'd', 2, [pixel_xres, pixel_yres], True))
        extra_tags.append((GTF_ModelTiepoint,
                           'd', 6, [0.0, 0.0, 0.0, origin_lon, origin_lat, 0.0], True))
        extra_tags.append((NINJO_TAGS["NTD_Magic"], 's', 0, "NINJO", True))
        extra_tags.append(
            (NINJO_TAGS["NTD_SatelliteNameID"], 'I', 1, sat_id, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_DateID"], 'I', 1, image_epoch, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_CreationDateID"], 'I', 1, file_epoch, True))
        extra_tags.append((NINJO_TAGS["NTD_ChannelID"], 'I', 1, chan_id, True))
        extra_tags.append((NINJO_TAGS["NTD_HeaderVersion"], 'i', 1, 2, True))
        if omit_filename_path:
            extra_tags.append((NINJO_TAGS["NTD_FileName"], 's', 0,
                               os.path.basename(output_fn), True))
        else:
            extra_tags.append(
                (NINJO_TAGS["NTD_FileName"], 's', 0, output_fn, True))
        extra_tags.append((NINJO_TAGS["NTD_DataType"], 's', 0, data_cat, True))
        # Hardcoded to 0
        extra_tags.append(
            (NINJO_TAGS["NTD_SatelliteNumber"], 's', 0, "\x00", True))

        if write_rgb:
            extra_tags.append((NINJO_TAGS["NTD_ColorDepth"], 'i', 1, 24, True))
        elif cmap:
            extra_tags.append((NINJO_TAGS["NTD_ColorDepth"], 'i', 1, 16, True))
        else:
            extra_tags.append((NINJO_TAGS["NTD_ColorDepth"], 'i', 1, 8, True))

        extra_tags.append(
            (NINJO_TAGS["NTD_DataSource"], 's', 0, data_source, True))
        extra_tags.append((NINJO_TAGS["NTD_XMinimum"], 'i', 1, 1, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_XMaximum"], 'i', 1, image_data.shape[1], True))
        extra_tags.append((NINJO_TAGS["NTD_YMinimum"], 'i', 1, 1, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_YMaximum"], 'i', 1, image_data.shape[0], True))
        extra_tags.append(
            (NINJO_TAGS["NTD_Projection"], 's', 0, projection, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_MeridianWest"], 'f', 1, meridian_west, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_MeridianEast"], 'f', 1, meridian_east, True))

        if radius_a is not None:
            extra_tags.append((NINJO_TAGS["NTD_EarthRadiusLarge"],
                               'f', 1, float(radius_a), True))
        if radius_b is not None:
            extra_tags.append((NINJO_TAGS["NTD_EarthRadiusSmall"],
                               'f', 1, float(radius_b), True))
        # extra_tags.append((NINJO_TAGS["NTD_GeodeticDate"], 's', 0, "\x00",
        # True)) # ---?
        if ref_lat1 is not None:
            extra_tags.append(
                (NINJO_TAGS["NTD_ReferenceLatitude1"], 'f', 1, ref_lat1, True))
        if ref_lat2 is not None:
            extra_tags.append(
                (NINJO_TAGS["NTD_ReferenceLatitude2"], 'f', 1, ref_lat2, True))
        if central_meridian is not None:
            extra_tags.append((NINJO_TAGS["NTD_CentralMeridian"],
                               'f', 1, central_meridian, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_PhysicValue"], 's', 0, physic_value, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_PhysicUnit"], 's', 0, physic_unit, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_MinGrayValue"], 'i', 1, min_gray_val, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_MaxGrayValue"], 'i', 1, max_gray_val, True))
        extra_tags.append((NINJO_TAGS["NTD_Gradient"], 'f', 1, gradient, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_AxisIntercept"], 'f', 1, axis_intercept, True))
        if altitude is not None:
            extra_tags.append(
                (NINJO_TAGS["NTD_Altitude"], 'f', 1, altitude, True))
        extra_tags.append((NINJO_TAGS["NTD_IsBlackLineCorrection"],
                           'i', 1, is_blac_corrected, True))
        extra_tags.append((NINJO_TAGS["NTD_IsAtmosphereCorrected"],
                           'i', 1, is_atmo_corrected, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_IsCalibrated"], 'i', 1, is_calibrated, True))
        extra_tags.append(
            (NINJO_TAGS["NTD_IsNormalized"], 'i', 1, is_normalized, True))
        extra_tags.append((NINJO_TAGS["NTD_TransparentPixel"],
                           'i', 1, transparent_pix, True))

        return args

    args = _create_args(image_data, pixel_xres, pixel_yres)

    tifargs = {}
    header_only_keys = ('byteorder', 'bigtiff', 'software', 'writeshape')
    for key in header_only_keys:
        if key in args:
            tifargs[key] = args[key]
            del args[key]

    if 'writeshape' not in args:
        args['writeshape'] = True
    if 'bigtiff' not in tifargs and \
            image_data.size * image_data.dtype.itemsize > 2000 * 2 ** 20:
        tifargs['bigtiff'] = True

    with tifffile.TiffWriter(output_fn, **tifargs) as tif:
        tif.save(image_data, **args)
        for _, scale in enumerate((2, 4, 8, 16)):
            shape = (image_data.shape[0] / scale,
                     image_data.shape[1] / scale)
            if shape[0] > tile_width and shape[1] > tile_length:
                args = _create_args(image_data[::scale, ::scale],
                                    pixel_xres * scale, pixel_yres * scale)
                for key in header_only_keys:
                    if key in args:
                        del args[key]
                tif.save(image_data[::scale, ::scale], **args)

    log.info("Successfully created a NinJo tiff file: '%s'" % (output_fn,))


# -----------------------------------------------------------------------------
#
# Read tags.
#
# -----------------------------------------------------------------------------
def read_tags(filename):
    """Will read tag, value pairs from Ninjo tiff file.

    :Parameters:
        filename : string
            Ninjo tiff file.

    :Return:
        A list tags, one tag dictionary per page.
    """
    pages = []
    with tifffile.TiffFile(filename) as tif:
        for page in tif:
            tags = {}
            for tag in page.tags.values():
                name, value = tag.name, tag.value
                try:
                    # Is it one of ours ?
                    name = int(name)
                    name = NINJO_TAGS_INV[name]
                except ValueError:
                    pass
                except KeyError:
                    name = tag.name
                tags[name] = value
            pages.append(tags)
    return pages

if __name__ == '__main__':
    import sys
    import getopt

    page_no = None
    print_color_maps = False
    opts, args = getopt.getopt(sys.argv[1:], "p:c")
    for key, val in opts:
        if key == "-p":
            page_no = int(val)
        if key == "-c":
            print_color_maps = True
    try:
        filename = args[0]
    except IndexError:
        print >> sys.stderr, """usage: python ninjotiff.py [<-p page-number>] [-c] <ninjotiff-filename>
    -p <page-number>: print page number (default are all pages).
    -c: print color maps (default is not to print color maps)."""
        sys.exit(2)

    pages = read_tags(filename)
    if page_no is not None:
        try:
            pages = [pages[page_no]]
        except IndexError:
            print >>sys.stderr, "Invalid page number '%d'" % page_no
            sys.exit(2)
    for page in pages:
        names = sorted(page.keys())
        print ""
        for name in names:
            if not print_color_maps and name == "color_map":
                continue
            print name, page[name]



from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
#import matplotlib as mpl
#mpl.use('Agg')
#from pylab import figure
from pylab import rand
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
from uuid import uuid4
import subprocess
from PIL import Image as PIL_Image
import numpy

def fig2data ( fig ):
    """
    @brief Convert a Matplotlib figure to a 4D numpy array with RGBA channels and return it
    @param fig a matplotlib figure
    @return a numpy 3D array of RGBA values
    """
    # draw the renderer
    fig.canvas.draw ( )

    # Get the RGBA buffer from the figure
    w,h = fig.canvas.get_width_height()
    buf = numpy.fromstring ( fig.canvas.tostring_argb(), dtype=numpy.uint8 )
    buf.shape = ( w, h, 4 )

    # canvas.tostring_argb give pixmap in ARGB mode. Roll the ALPHA channel to have it in RGBA mode
    buf = numpy.roll ( buf, 3, axis = 2 )
    return buf

def fig2img ( fig ):
    """
    @brief Convert a Matplotlib figure to a PIL Image in RGBA format and return it
    @param fig a matplotlib figure
    @return a Python Imaging Library ( PIL ) image
    """
    # put the figure pixmap into a numpy array
    buf = fig2data ( fig )
    w, h, d = buf.shape
    return PIL_Image.frombytes( "RGBA", ( w ,h ), buf.tostring( ) )


def TRTimage( TRTcell_IDs, TRTcells, obj_area, minRank=8, alpha_max=1.0, plot_vel=True):

    # define size of image 
    nx = obj_area.x_size
    ny = obj_area.y_size

    # create new figure 
    fig = Figure()
    # canvas figure 
    canvas = FigureCanvas(fig)
    # get dots per inch of the screen
    DPI = fig.get_dpi()
    # print "DPI", DPI
    fig.set_size_inches(nx/float(DPI),ny/float(DPI))
    # get axis object 
    ax = fig.add_subplot(111, aspect='equal')
    ## eliminates margins totally 
    fig.subplots_adjust(left=0.0,right=1.0,bottom=0.0,top=1.0, wspace=0, hspace=0)
    #plt.subplots_adjust(left=0, right=1, bottom=0, top=1, wspace=0, hspace=0) # does only work with x11 display 
    # set limits of the axis
    ax.set_xlim(0, nx)
    ax.set_ylim(0, ny)
    # set transparent backgroud 
    fig.patch.set_alpha(0.0)        # transparent outside of diagram  
    ax.set_axis_bgcolor([1,0,0,0])  # transparent color inside diagram 

    # define arrow properties 
    head_width  = 0.006 * min(obj_area.x_size,obj_area.x_size)
    head_length = 2 * head_width

    pixel_size_x_km = 0.001 * obj_area.pixel_size_x
    pixel_size_y_km = 0.001 * obj_area.pixel_size_y

    for cell in TRTcell_IDs:

        if TRTcells[cell].RANKr > minRank:

            (x0,y0) = obj_area.get_xy_from_lonlat(TRTcells[cell].lon, TRTcells[cell].lat, outside_error=False, return_int=False)
            y0 = (obj_area.y_size-1)-y0
            # print (x0,y0)

            vx = TRTcells[cell].vel_x
            vy = TRTcells[cell].vel_y
   
            # !!!scaling of width and height is not correct, that is on map projection, but not on the ground!!!
            e = Ellipse( xy     =  (x0, y0),                           \
                         width  =  2*TRTcells[cell].ell_S / pixel_size_x_km, \
                         height =  2*TRTcells[cell].ell_L / pixel_size_y_km, \
                         angle  = -TRTcells[cell].angle )
            
            ax.add_artist(e)
            e.set_clip_box(ax.bbox)
            
            if TRTcells[cell].RANKr <= 12:
                cell_color="white"
                alpha = (alpha_max-0.2) / 12. * TRTcells[cell].RANKr
            elif TRTcells[cell].RANKr <= 15:
                cell_color="white"
                alpha = alpha_max
            elif TRTcells[cell].RANKr <= 25:
                cell_color="green"
                alpha = alpha_max
            elif TRTcells[cell].RANKr <= 35:
                cell_color="yellow"
                alpha = alpha_max
            else:
                cell_color="red"
                alpha = alpha_max
            # print "cell ID: %s, cell rank: %2d, cell_color:%7s, alpha = %4.1f" % (cell, TRTcells[cell].RANKr, cell_color, alpha)
            e.set_alpha(alpha)       # transparency: 0.0 transparent, 1 total visible  
            e.set_facecolor(cell_color)  # "white" or [1,1,1]

            if plot_vel:
                ax.arrow(x0, y0, vx, vy, head_width = head_width, head_length = head_length, fc=cell_color, ec=cell_color)

    if 1==1:
        # print " !!! convert fig to image by function fig2img !!!"
        ### this would avoid saving into a file, but it fills the transparent areas with "white"
        PIL_image = fig2img ( fig )  
    else: 
        tmp_file = '/tmp/TRT_'+str(uuid4())+'.png'
        # print tmp_file
        plt.savefig(tmp_file, dpi=DPI, transparent=True) #, bbox_inches='tight'
        # subprocess.call("display "+tmp_file+" &", shell=True) 
        PIL_image = PIL_Image.open(tmp_file)
        subprocess.call("rm "+tmp_file+" &", shell=True) 

    return PIL_image



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 Abhay Devasthale and Martin Raspaud

# Author(s):

#   Abhay Devasthale <abhay.devasthale@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Sajid Pareeth <sajid.pareeth@fmch.it>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Read a gac file.
Reads L1b GAC data from KLM series of satellites (NOAA-15 and later) and does most of the computations.
Format specification can be found here:
http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec83142-1.htm

"""

import os
import glob
from ConfigParser import ConfigParser
import logging

import numpy as np
from pygac.lac_klm import KLMReader
from pygac.lac_pod import PODReader
from mpop import CONFIG_PATH

LOGGER = logging.getLogger(__name__)


def load(satscene, *args, **kwargs):
    """Read data from file and load it into *satscene*.
    A possible *calibrate* keyword argument is passed to the AAPP reader. 
    Should be 0 for off (counts), 1 for default (brightness temperatures and
    reflectances), and 2 for radiances only.
    """
    del args

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2",
                                    raw=True):
        options[option] = value

    if kwargs.get("filename") is not None:
        options["filename"] = kwargs["filename"]
        options["dir"] = None

    options["calibrate"] = kwargs.get("calibrate", True)

    LOGGER.info("Loading instrument '%s'" % satscene.instrument_name)

    try:
        CASES[satscene.instrument_name](satscene, options)
    except KeyError:
        raise KeyError("Unknown instrument '%s'" % satscene.instrument_name)


def load_avhrr(satscene, options):
    """Read avhrr data from file and load it into *satscene*.
    """
    if "filename" not in options:
        raise IOError("No filename given, cannot load.")

    values = {"orbit":      satscene.orbit,
              "satname":    satscene.satname,
              "number":     satscene.number,
              "instrument": satscene.instrument_name,
              "satellite":  satscene.fullname
              }

    if options["dir"] is None:
        filename = options["filename"]
    else:
        filename = os.path.join(satscene.time_slot.strftime(options["dir"]) % values,
                                satscene.time_slot.strftime(
                                    options["filename"])
                                % values)

        file_list = glob.glob(filename)

        if len(file_list) > 1:
            raise IOError("More than one l1b file matching!")
        elif len(file_list) == 0:
            raise IOError("No l1b file matching!: " + filename)

        filename = file_list[0]

    LOGGER.debug("Loading from " + filename)

    with open(filename) as fdes:
        data = fdes.read(3)
    if data in ["CMS", "NSS", "UKM", "DSS"]:
        reader = KLMReader
        chn_dict = AVHRR3_CHANNEL_NAMES
    else:
        reader = PODReader
        chn_dict = AVHRR_CHANNEL_NAMES


    chns = satscene.channels_to_load & set(chn_dict.keys())
    LOGGER.info("Loading channels " + str(sorted(list(chns))))

    if len(chns) == 0:
        return

    scene = reader()
    scene.read(filename)
    scene.get_lonlat()
    scene.adjust_clock_drift()
    channels = scene.get_calibrated_channels()

    # scene.navigate()
    try:
        from pyresample import geometry
    except ImportError as ex_:

        LOGGER.debug("Could not load pyresample: " + str(ex_))

        satscene.lat = scene.lats
        satscene.lon = scene.lons
    else:
        satscene.area = geometry.SwathDefinition(lons=scene.lons,
                                                 lats=scene.lats)
        area_name = ("swath_" + satscene.fullname + "_" +
                     str(satscene.time_slot) + "_"
                     + str(scene.lats.shape))
        satscene.area.area_id = area_name
        satscene.area.name = "Satellite projection"
        satscene.area_id = area_name

    for chn in chns:
        data = channels[:, :, chn_dict[chn]]
        if np.ma.count(data) > 0:
            satscene[chn].data = np.ma.masked_invalid(data, copy=False)
            satscene[chn].area = satscene.area


AVHRR3_CHANNEL_NAMES = {"1": 0, "2": 1, "3A": 2, "3B": 3, "4": 4, "5": 5}
AVHRR_CHANNEL_NAMES = {"1": 0, "2": 1, "3": 2, "4": 3, "5": 4}

CASES = {
    "avhrr/1": load_avhrr,
    "avhrr/2": load_avhrr,
    "avhrr/3": load_avhrr,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009, 2013, 2015, 2016.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of the mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Palette holder module.
"""

import numpy as np


def tv_legend():
    """Palette for TV.
    """
    legend = []
    legend.append((0,   0,   0))  # Unprocessed: Black
    legend.append((0, 120,   0))  # Land
    legend.append((0,   0, 215))  # Sea: Blue
    legend.append((0, 120,   0))  # Land (Snow on land)
    legend.append((0,   0, 215))  # Sea: Blue (Snow/Ice on sea)

    for i in range(5, 256):
        # All other pixel values are grey according to IR temp.
        legend.append((i, i, i))

    return convert_palette(legend)


def vv_legend():
    """Palette for Swedish road authorities (Vägverket).
    """
    legend = []
    legend.append((0,   0,   0))  # Unprocessed: Black
    legend.append((0, 120,   0))  # Land
    legend.append((0,   0, 215))  # Sea: Blue
    # Cloud type values 5 to 8:
    legend.append((255, 150,   0))  # Very low cumuliform
    legend.append((255, 100,   0))  # Very low
    legend.append((255, 220,   0))  # Low cumuliform
    legend.append((255, 180,   0))  # Low

    for i in range(7, 256):
        # All other pixel values are grey according to IR temp.
        legend.append((i, i, i))

    return convert_palette(legend)


def cloud_phase():
    """Palette for cloud phase.
    """
    legend = []
    legend.append((0,    0,   0))  # Unprocessed: Black
    legend.append((0,    0, 215))  # Water Clouds: Blue
    legend.append((240, 240, 240))  # Ice Clouds: Almost White
    legend.append((120, 120,   0))  # Uncertain Phase: ?

    return convert_palette(legend)


def cms_modified():
    """Palette for regular cloud classification.
    """
    return nwcsaf_cloudtype()


def nwcsaf_cloudtype():
    """Palette for regular cloud classification.
    """
    legend = []
    legend.append((100, 100, 100))  # Unprocessed: Grey
    legend.append((0, 120,   0))
    legend.append((0,   0,   0))  # Sea: Black
    legend.append((250, 190, 250))  # Snow
    legend.append((220, 160, 220))  # Sea-ice

    legend.append((255, 150,   0))  # Very low cumuliform
    legend.append((255, 100,   0))  # Very low
    legend.append((255, 220,   0))  # Low cumuliform
    legend.append((255, 180,   0))  # Low
    legend.append((255, 255, 140))  # Medium cumuliform
    legend.append((240, 240,   0))  # Medium
    legend.append((250, 240, 200))  # High cumiliform
    legend.append((215, 215, 150))  # High
    legend.append((255, 255, 255))  # Very high cumuliform
    legend.append((230, 230, 230))  # Very high

    legend.append((0,  80, 215))  # Semi-transparent thin
    legend.append((0, 180, 230))  # Semi-transparent medium
    legend.append((0, 240, 240))  # Semi-transparent thick
    legend.append((90, 200, 160))  # Semi-transparent above
    legend.append((200,   0, 200))  # Broken
    legend.append((95,  60,  30))  # Undefined: Brown

    return convert_palette(legend)


def ctth_height():
    """CTTH height palette.
    """
    legend = []
    legend.append((0,   0,   0))
    legend.append((255,   0, 216))  # 0 meters
    legend.append((126,   0,  43))
    legend.append((153,  20,  47))
    legend.append((178,  51,   0))
    legend.append((255,  76,   0))
    legend.append((255, 102,   0))
    legend.append((255, 164,   0))
    legend.append((255, 216,   0))
    legend.append((216, 255,   0))
    legend.append((178, 255,   0))
    legend.append((153, 255,   0))
    legend.append((0, 255,   0))
    legend.append((0, 140,  48))
    legend.append((0, 178, 255))
    legend.append((0, 216, 255))
    legend.append((0, 255, 255))
    legend.append((238, 214, 210))
    legend.append((239, 239, 223))
    legend.append((255, 255, 255))  # 10,000 meters
    for i in range(79):
        legend.append((255, 255, 255))
    legend.append((224, 224, 224))

    return convert_palette(legend)


def ctth_height_pps():
    """CTTH height palette for NWCSAF/PPS.
    Identical to the one found in the hdf5 files.
    """
    legend = []
    legend.append((255,   0, 216))  # 0 meters
    legend.append((255,   0, 216))  # 0 meters
    legend.append((255,   0, 216))  # 0 meters
    legend.append((126,   0,  43))
    legend.append((126,   0,  43))
    legend.append((153,  20,  47))
    legend.append((153,  20,  47))
    legend.append((153,  20,  47))
    legend.append((178,  51,   0))
    legend.append((178,  51,   0))
    legend.append((255,  76,   0))
    legend.append((255,  76,   0))
    legend.append((255,  76,   0))
    legend.append((255, 102,   0))
    legend.append((255, 102,   0))
    legend.append((255, 164,   0))
    legend.append((255, 164,   0))
    legend.append((255, 164,   0))
    legend.append((255, 216,   0))
    legend.append((255, 216,   0))
    legend.append((216, 255,   0))
    legend.append((216, 255,   0))
    legend.append((178, 255,   0))
    legend.append((178, 255,   0))
    legend.append((178, 255,   0))
    legend.append((153, 255,   0))
    legend.append((153, 255,   0))
    legend.append((0, 255,   0))
    legend.append((0, 255,   0))
    legend.append((0, 255,   0))
    legend.append((0, 140,  48))
    legend.append((0, 140,  48))
    legend.append((0, 178, 255))
    legend.append((0, 178, 255))
    legend.append((0, 178, 255))
    legend.append((0, 216, 255))
    legend.append((0, 216, 255))
    legend.append((0, 255, 255))
    legend.append((0, 255, 255))
    legend.append((0, 255, 255))
    legend.append((238, 214, 210))
    legend.append((238, 214, 210))
    legend.append((239, 239, 223))
    legend.append((239, 239, 223))
    for idx in range(47, 150):
        legend.append((255, 255, 255))  # 10,000 meters

    for idx in range(150, 256):
        legend.append((0, 0, 0))

    return convert_palette(legend)


def chlorophyll_a():
    """Chlorophyll-A legend for MERIS"""
    raise NotImplementedError("This palette is not implemented - " +
                              "it was earlier though...")


# --------------------------------------------
#   Define colour palette LUT for SST palette image
#   colour shading; blue(cold) -> green -> yellow -> red(warm)
#


def sstlut_osisaf_metno():

    legend = []
    legend.append((0, 0, 0))  # /*  Black (out of image)  */
    legend.append((82, 82, 82))  # /*  Dark grey (land)  */
    legend.append((187, 187, 187))  # /* Light grey (cloud contaminated)  */
    legend.append((255, 255, 255))  # /* White (sea ice and snow)  */

    legend.append((255, 0, 255))  # /* Starting at 4 = pink */
    legend.append((195, 0, 129))  # /* Dark pink */
    legend.append((129, 0, 47))  # /* Dark red */
    legend.append((195, 0, 0))  # /* Medium dark red */

    r = 255  # /* Red */
    g = 0
    b = 0
    for i in range(8, 11, 1):
        legend.append((r, g, b))
        r = r - 19
        g = g + 43

    r = 200  # /* Brown */
    g = 128
    b = 0
    for i in range(11, 16, 1):
        legend.append((r, g, b))
        r = r + 11
        g = g + 26
        b = b + 13

    r = 65535 / 256  # /* Yellow */
    g = 65535 / 256
    b = 16185 / 256
    legend.append((r, g, b))
    legend.append((52000 / 256, 65535 / 256, 13500 / 256))
    legend.append((35000 / 256, 65535 / 256, 7000 / 256))

    r = 0  # /* Green */
    g = 65535 / 256
    b = 0
    for i in range(19, 22, 1):
        legend.append((r, g, b))
        g = g - 12422 / 256

    r = 0  # /* Dark Green */
    g = 28269 / 256
    b = 0
    for i in range(22, 26, 1):
        legend.append((r, g, b))
        g = g - 7067 / 256
        b = b + 16383 / 256

    r = 0
    g = 0
    b = 65535 / 256
    legend.append((r, g, b))  # Blue
    legend.append((25700 / 256, 25700 / 256, 65535 / 256))  # Dark purple
    legend.append((38550 / 256, 38550 / 256, 65535 / 256))  # Light purple

    for i in range(29, 256):
        legend.append((0, 0, 0))

    return convert_palette(legend)


def convert_palette(palette):
    """Convert palette from [0,255] range to [0,1].
    """
    new_palette = []
    for i in palette:
        new_palette.append((i[0] / 255.0,
                            i[1] / 255.0,
                            i[2] / 255.0))
    return new_palette


def convert_palette2colormap(palette):
    """Convert palette from [0,255] range to [0,1].
    """
    from trollimage.colormap import Colormap
    j = 0
    n_pal = len(palette) - 1
    values = []
    colors = []

    red = [r for (r, g, b) in palette]
    green = [g for (r, g, b) in palette]
    blue = [b for (r, g, b) in palette]

    max_pal = max(max(red), max(blue), max(green))
    if max_pal <= 1.0:
        # print "palette already normalized"
        denom = 1.0
    else:
        # print "palette normalized to 255"
        denom = 255.0

    for i in palette:
        values.append((n_pal - j) / float(n_pal))
        colors.append((i[0] / denom, i[1] / denom, i[2] / denom))
        j = j + 1
    # reverse order to the entries
    values = values[::-1]
    colors = colors[::-1]

    # for i in palette:
    #    values.append( j /  float(n_pal))
    #    colors.append((i[0] / 255.0, i[1] / 255.0, i[2] / 255.0))
    #    j=j+1

    # attention:
    # Colormap(values, colors) uses the second input option of Colormap
    # values has to be a list (not a tuple) and
    # colors has to be the corresponding list of color tuples

    return Colormap(values, colors)


class LogColors(object):

    """
    Defines colors to use with `logdata2image`

    """

    def __init__(self, nodata, zeros, over, breaks):
        self.nodata = nodata
        self.zeros = zeros
        self.over = over
        self.breaks = breaks

    def palette(self, N=256):
        """
        Build a palette for logarithmic data images.

        """

        max_value = self.breaks[-1][0]

        pal = np.zeros((N, 3), dtype=np.uint8)

        b_last, rgb_last = self.breaks[0]
        for b, rgb in self.breaks[1:]:
            # Get a slice of the palette array for the current interval
            p = pal[
                np.log(b_last + 1) * N / np.log(max_value):np.log(b + 1) * N / np.log(max_value)]
            for i in range(3):  # red, green, blue
                p[:, i] = np.linspace(rgb_last[i], rgb[i], p.shape[0])
            b_last = b
            rgb_last = rgb

        pal[0] = self.nodata
        pal[1] = self.zeros
        pal[-1] = self.over

        return pal


class TriColors(LogColors):

    """
    Use three color tones in the intervals between the elements of *breaks*.

    """
    color_tones = [((0, 0, 200), (150, 150, 255)),  # dark to light blue
                   ((150, 150, 0), (255, 255, 8)),  # greyish to bright yellow
                   ((230, 150, 100), (230, 0, 0))]  # green to red

    nodata = (0, 0, 0)  # black
    # zeros = (20, 0, 20) # dark purple
    # black  #There is no need to mark zeros with another col
    zeros = (0, 0, 0)
    over = (255, 0, 0)  # bright red

    def __init__(self, breaks):
        breaks = [(breaks[0], TriColors.color_tones[0][0]),
                  (breaks[1], TriColors.color_tones[0][1]),

                  (breaks[1], TriColors.color_tones[1][0]),
                  (breaks[2], TriColors.color_tones[1][1]),

                  (breaks[2], TriColors.color_tones[2][0]),
                  (breaks[3], TriColors.color_tones[2][1])]

        LogColors.__init__(self, TriColors.nodata, TriColors.zeros,
                           TriColors.over, breaks)

CPP_COLORS = {'cpp_cot': TriColors([0, 3.6, 23, 700]),  # ISCCP intervals
              'cpp_reff': TriColors([0, 10, 20, 1000])}

CPP_COLORS['cot'] = CPP_COLORS['cpp_cot']
CPP_COLORS['reff'] = CPP_COLORS['cpp_reff']


def get_ctp_legend():
    """
    Get the Cloud Top Pressure color palette
    """

    legend = []
    legend.append((0, 0, 0))     # No data
    legend.append((255, 0, 216))  # 0: 1000-1050 hPa (=100000-105000 Pa)
    legend.append((126, 0, 43))  # 1: 950-1000 hPa
    legend.append((153, 20, 47))  # 2: 900-950 hPa
    legend.append((178, 51, 0))  # 3: 850-900 hPa
    legend.append((255, 76, 0))  # 4: 800-850 hPa
    legend.append((255, 102, 0))  # 5: 750-800 hPa
    legend.append((255, 164, 0))  # 6: 700-750 hPa
    legend.append((255, 216, 0))  # 7: 650-700 hPa
    legend.append((216, 255, 0))  # 8: 600-650 hPa
    legend.append((178, 255, 0))  # 9: 550-600 hPa
    legend.append((153, 255, 0))  # 10: 500-550 hPa
    legend.append((0, 255, 0))   # 11: 450-500 hPa
    legend.append((0, 140, 48))  # 12: 400-450 hPa
    legend.append((0, 178, 255))  # 13: 350-400 hPa
    legend.append((0, 216, 255))  # 14: 300-350 hPa
    legend.append((0, 255, 255))  # 15: 250-300 hPa
    legend.append((238, 214, 210))  # 16: 200-250 hPa
    legend.append((239, 239, 223))  # 17: 150-200 hPa
    legend.append((255, 255, 255))  # 18: 100-150 hPa
    legend.append((255, 255, 255))  # 19: 50-100 hPa
    legend.append((255, 255, 255))  # 20: 0-50 hPa  (=0-5000 Pa)

    palette = convert_palette(legend)
    return palette


def get_reff_legend():
    return get_log_legend('reff')


def get_cot_legend():
    return get_log_legend('cot')


def get_log_legend(product_name):
    # This is the same data as is used in logdata2image (when indata as for
    # the calls from cppimage)
    legend = CPP_COLORS[product_name].palette()
    palette = convert_palette(legend)
    return palette


def oca_get_scenetype_legend():

    # Colorize using PPS/CPP palette
    legend = np.array([[170, 130, 255],  # purple/blue for liquid (cph == 1)
                       [220, 200, 255],  # almost white for ice (cph == 2)
                       [255, 200, 200]   # Redish for multi layer clouds
                       ])
    legend = np.vstack([np.zeros((111, 3)), legend])
    palette = convert_palette(legend)
    return palette



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2013, 2014.

# Author(s):

#   Lars Ørum Rasmussen <ras@dmi.dk>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.
#
"""This modules describes the sarx instrument from the TerraSAR-X satellite.
"""
import numpy as np

from mpop.compositer import Compositer
import logging

LOG = logging.getLogger(__name__)


class SarxCompositer(Compositer):

    """This class sets up the SAR-X instrument channel list.
    """

    instrument_name = "sarx"

    def average(self, downscaling_factor=2, average_window=None):
        """
        Makes a mean convolution of an image.

        :Parameters:
         `downscaling_factor` : int
             image downscaling factor, default is a factor 2.
         `average_window` : int
             window size for calculating mean values, default is
             the same as downscaling_factor.

        :Returns:
         `image` : GeoImage
             mean convoluted image.
        """
        from mpop.imageo.geo_image import GeoImage
        from pyresample import geometry
        import scipy.ndimage as ndi

        self.check_channels(9.65)

        if average_window == None:
            average_window = downscaling_factor

        LOG.info("Downsampling a factor %d and averaging " % downscaling_factor +
                 "in a window of %dx%d" % (average_window, average_window))

        ch = self[9.65]

        # If average window and downscale factor is the same
        # the following could be used:
        #
        #    data = data.reshape([shight, hight/shight,
        #                         swidth, width/swidth]).mean(3).mean(1)

        # avg kernel
        kernel = (np.ones((average_window, average_window), dtype=np.float)
                  / (average_window * average_window))
        # do convolution
        data = ndi.filters.correlate(ch.data.astype(np.float), kernel,
                                     mode='nearest')
        # downscale
        data = data[1::downscaling_factor, 1::downscaling_factor]

        # New area, and correct for integer truncation.
        p_size_x, p_size_y = (ch.area.pixel_size_x * downscaling_factor,
                              ch.area.pixel_size_y * downscaling_factor)
        area_extent = (ch.area.area_extent[0],
                       ch.area.area_extent[1],
                       ch.area.area_extent[0] + data.shape[1] * p_size_x,
                       ch.area.area_extent[1] + data.shape[0] * p_size_y)

        area = geometry.AreaDefinition(self._data_holder.satname +
                                       self._data_holder.instrument_name +
                                       str(area_extent) +
                                       str(data.shape),
                                       "On-the-fly area",
                                       ch.area.proj_id, ch.area.proj_dict,
                                       data.shape[1], data.shape[0],
                                       area_extent)

        return GeoImage(data, area, self.time_slot,
                        fill_value=(0,), mode='L')

    average.prerequisites = set([9.65, ])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2012, 2014, 2015.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin for reading PPS's cloud products hdf files.
"""
import ConfigParser
import os.path
from mpop import CONFIG_PATH
import mpop.channel
import numpy as np
import pyresample.utils

import glob
from mpop.utils import get_logger
from mpop.projector import get_area_def
from datetime import datetime

LOG = get_logger('satin/msg_hdf')
COMPRESS_LVL = 6

ctype_lut = ['0: Not processed',
             '1: Cloud free land',
             '2: Cloud free sea',
             '3: Snow/ice contaminated land',
             '4: Snow/ice contaminated sea',
             '5: Very low cumiliform cloud',
             '6: Very low stratiform cloud',
             '7: Low cumiliform cloud',
             '8: Low stratiform cloud',
             '9: Medium level cumiliform cloud',
             '10: Medium level stratiform cloud',
             '11: High and opaque cumiliform cloud',
             '12: High and opaque stratiform cloud',
             '13:Very high and opaque cumiliform cloud',
             '14: Very high and opaque stratiform cloud',
             '15: Very thin cirrus cloud',
             '16: Thin cirrus cloud',
             '17: Thick cirrus cloud',
             '18: Cirrus above low or medium level cloud',
             '19: Fractional or sub-pixel cloud',
             '20: Undefined']
phase_lut = ['1: Not processed or undefined',
             '2: Water',
             '4: Ice',
             '8: Tb11 below 260K',
             '16: value not defined',
             '32: value not defined',
             '64: value not defined',
             '128: value not defined']
quality_lut = ['1: Land',
               '2: Coast',
               '4: Night',
               '8: Twilight',
               '16: Sunglint',
               '32: High terrain',
               '64: Low level inversion',
               '128: Nwp data present',
               '256: Avhrr channel missing',
               '512: Low quality',
               '1024: Reclassified after spatial smoothing',
               '2048: Stratiform-Cumuliform Distinction performed',
               '4096: bit not defined',
               '8192: bit not defined',
               '16384: bit not defined',
               '32768: bit not defined']
processing_flags_lut = ["1: Not processed",
                        "2: Cloudy",
                        "4: Opaque cloud",
                        "8: RTTOV IR simulations available",
                        "16: Missing NWP data",
                        "32: thermal inversion avaliable",
                        "64: Missing AVHRR data",
                        "128: RTTOV IR simulation applied",
                        "256: Windowing technique applied",
                        "512: bit not defined",
                        "1024: bit not defined",
                        "2048: bit not defined",
                        "4096: bit not defined",
                        "8192: bit not defined",
                        "16384: Quality estimation avaliable",
                        "32768: Low confidence"]


class UnknownChannelError(Exception):
    pass


def pcs_def_from_region(region):
    items = region.proj_dict.items()
    return ' '.join([t[0] + '=' + t[1] for t in items])


def _get_area_extent(cfac, lfac, coff, loff, numcols, numlines):
    """Get the area extent from msg parameters.
    """

    xur = (numcols - coff) * 2 ** 16 / (cfac * 1.0)
    xur = np.deg2rad(xur) * 35785831.0
    xll = (-1 - coff) * 2 ** 16 / (cfac * 1.0)
    xll = np.deg2rad(xll) * 35785831.0
    xres = (xur - xll) / numcols
    xur, xll = xur - xres / 2, xll + xres / 2
    yll = (numlines - loff) * 2 ** 16 / (-lfac * 1.0)
    yll = np.deg2rad(yll) * 35785831.0
    yur = (-1 - loff) * 2 ** 16 / (-lfac * 1.0)
    yur = np.deg2rad(yur) * 35785831.0
    yres = (yur - yll) / numlines
    yll, yur = yll + yres / 2, yur - yres / 2
    return xll, yll, xur, yur


def get_area_extent(filename):
    """Get the area extent of the data in *filename*.
    """
    import h5py
    h5f = h5py.File(filename, 'r')
    aex = _get_area_extent(h5f.attrs["CFAC"],
                           h5f.attrs["LFAC"],
                           h5f.attrs["COFF"],
                           h5f.attrs["LOFF"],
                           h5f.attrs["NC"],
                           h5f.attrs["NL"])
    h5f.close()
    return aex


def _get_palette(h5f, dsname):
    try:
        p = h5f[dsname].attrs['PALETTE']
        return h5f[p].value
    except KeyError:
        return None


class InfoObject(object):

    """Simple data and info container.
    """

    def __init__(self):
        self.info = {}
        self.data = None

# ----------------------------------------


class MsgCloudTypeData(object):

    """NWCSAF/MSG Cloud Type data layer
    """

    def __init__(self):
        self.data = None
        self.scaling_factor = 1
        self.offset = 0
        self.num_of_lines = 0
        self.num_of_columns = 0
        self.product = ""
        self.id = ""


class MsgCloudType(mpop.channel.GenericChannel):

    """NWCSAF/MSG Cloud Type data structure as retrieved from HDF5
    file. Resolution sets the nominal resolution of the data.
    """

    def __init__(self):
        mpop.channel.GenericChannel.__init__(self, "CloudType")
        self.filled = False
        self.name = "CloudType"
        self.package = ""
        self.saf = ""
        self.product_name = ""
        self.num_of_columns = 0
        self.num_of_lines = 0
        self.projection_name = ""
        self.pcs_def = ""
        self.xscale = 0
        self.yscale = 0
        self.ll_lon = 0.0
        self.ll_lat = 0.0
        self.ur_lon = 0.0
        self.ur_lat = 0.0
        self.region_name = ""
        self.cfac = 0
        self.lfac = 0
        self.coff = 0
        self.loff = 0
        self.nb_param = 0
        self.gp_sc_id = 0
        self.image_acquisition_time = 0
        self.spectral_channel_id = 0
        self.nominal_product_time = 0
        self.sgs_product_quality = 0
        self.sgs_product_completeness = 0
        self.product_algorithm_version = ""
        self.cloudtype = None
        self.processing_flags = None
        self.cloudphase = None
        self.shape = None
        self.satid = ""
        self.qc_straylight = -1
        self.cloudtype_palette = None
        self.cloudphase_palette = None

    def __str__(self):
        try:
            shape = self.cloudtype.shape
        except AttributeError:
            shape = self.shape

        return ("'%s: shape %s, resolution %sm'" %
                (self.name,
                 shape,
                 self.resolution))

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return self.filled

# ------------------------------------------------------------------
    def read(self, filename):
        """Reader for the NWCSAF/MSG cloudtype. Use *filename* to read data.
        """
        import h5py

        self.cloudtype = MsgCloudTypeData()
        self.processing_flags = MsgCloudTypeData()
        self.cloudphase = MsgCloudTypeData()

        LOG.debug("Filename = <" + str(filename) + ">")
        h5f = h5py.File(filename, 'r')
        # pylint: disable-msg=W0212
        self.package = h5f.attrs["PACKAGE"]
        self.saf = h5f.attrs["SAF"]
        self.product_name = h5f.attrs["PRODUCT_NAME"]
        self.num_of_columns = h5f.attrs["NC"]
        self.num_of_lines = h5f.attrs["NL"]
        self.projection_name = h5f.attrs["PROJECTION_NAME"]
        self.region_name = h5f.attrs["REGION_NAME"]
        self.cfac = h5f.attrs["CFAC"]
        self.lfac = h5f.attrs["LFAC"]
        self.coff = h5f.attrs["COFF"]
        self.loff = h5f.attrs["LOFF"]
        self.nb_param = h5f.attrs["NB_PARAMETERS"]
        self.gp_sc_id = h5f.attrs["GP_SC_ID"]
        self.image_acquisition_time = h5f.attrs["IMAGE_ACQUISITION_TIME"]
        self.spectral_channel_id = h5f.attrs["SPECTRAL_CHANNEL_ID"]
        self.nominal_product_time = datetime.strptime(h5f.attrs["NOMINAL_PRODUCT_TIME"],
                                                      "%Y%m%d%H%M")
        self.sgs_product_quality = h5f.attrs["SGS_PRODUCT_QUALITY"]
        self.sgs_product_completeness = h5f.attrs["SGS_PRODUCT_COMPLETENESS"]
        self.product_algorithm_version = h5f.attrs[
            "PACKAGE"] + h5f.attrs["PRODUCT_NAME"] + h5f.attrs["PRODUCT_ALGORITHM_VERSION"]
        # pylint: enable-msg=W0212
        # ------------------------

        # The cloudtype data
        if 'CT' not in h5f.keys():
            raise IOError("No field CT in product " +
                          str(self.product_name) +
                          "\n\tPerhaps you have found the wrong file for this product?")
        h5d = h5f['CT']
        self.cloudtype.data = h5d[:, :]
        self.cloudtype.scaling_factor = h5d.attrs["SCALING_FACTOR"]
        self.cloudtype.offset = h5d.attrs["OFFSET"]
        self.cloudtype.num_of_lines = h5d.attrs["N_LINES"]
        self.cloudtype.num_of_columns = h5d.attrs["N_COLS"]
        self.shape = (self.cloudtype.num_of_lines,
                      self.cloudtype.num_of_columns)
        self.cloudtype.product = h5d.attrs["PRODUCT"]
        self.cloudtype.id = h5d.attrs["ID"]
        self.cloudtype_palette = _get_palette(h5f, 'CT')
        # ------------------------

        # The cloud phase data
        h5d = h5f['CT_PHASE']
        self.cloudphase.data = h5d[:, :]
        self.cloudphase.scaling_factor = h5d.attrs["SCALING_FACTOR"]
        self.cloudphase.offset = h5d.attrs["OFFSET"]
        self.cloudphase.num_of_lines = h5d.attrs["N_LINES"]
        self.cloudphase.num_of_columns = h5d.attrs["N_COLS"]
        self.cloudphase.product = h5d.attrs["PRODUCT"]
        self.cloudphase.id = h5d.attrs["ID"]
        self.cloudphase_palette = _get_palette(h5f, 'CT_PHASE')

        # ------------------------

        # The cloudtype processing/quality flags
        h5d = h5f['CT_QUALITY']
        self.processing_flags.data = h5d[:, :]
        self.processing_flags.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.processing_flags.offset = h5d.attrs["OFFSET"]
        self.processing_flags.num_of_lines = h5d.attrs["N_LINES"]
        self.processing_flags.num_of_columns = h5d.attrs["N_COLS"]
        self.processing_flags.product = h5d.attrs["PRODUCT"]
        self.processing_flags.id = h5d.attrs["ID"]
        # ------------------------

        h5f.close()

        self.cloudtype = self.cloudtype.data
        self.cloudphase = self.cloudphase.data
        self.processing_flags = self.processing_flags.data

        self.area = get_area_from_file(filename)

        self.filled = True

    def save(self, filename, **kwargs):
        """Save the current cloudtype object to hdf *filename*, in pps format.
        """
        import h5py
        ctype = self.convert2pps()
        LOG.info("Saving CType hdf file...")
        ctype.save(filename, **kwargs)
        h5f = h5py.File(filename, mode="a")
        h5f.attrs["straylight_contaminated"] = self.qc_straylight
        h5f.close()
        LOG.info("Saving CType hdf file done !")

    def project(self, coverage):
        """Remaps the NWCSAF/MSG Cloud Type to cartographic map-projection on
        area give by a pre-registered area-id. Faster version of msg_remap!
        """
        LOG.info("Projecting channel %s..." % (self.name))

        region = coverage.out_area
        dest_area = region.area_id

        retv = MsgCloudType()

        retv.name = self.name
        retv.package = self.package
        retv.saf = self.saf
        retv.product_name = self.product_name
        retv.region_name = dest_area
        retv.cfac = self.cfac
        retv.lfac = self.lfac
        retv.coff = self.coff
        retv.loff = self.loff
        retv.nb_param = self.nb_param
        retv.gp_sc_id = self.gp_sc_id
        retv.image_acquisition_time = self.image_acquisition_time
        retv.spectral_channel_id = self.spectral_channel_id
        retv.nominal_product_time = self.nominal_product_time
        retv.sgs_product_quality = self.sgs_product_quality
        retv.sgs_product_completeness = self.sgs_product_completeness
        retv.product_algorithm_version = self.product_algorithm_version

        retv.cloudtype = coverage.project_array(self.cloudtype)
        retv.cloudtype_palette = self.cloudtype_palette

        retv.cloudphase = coverage.project_array(self.cloudphase)
        retv.cloudphase_palette = self.cloudphase_palette

        retv.processing_flags = \
            coverage.project_array(self.processing_flags)

        retv.qc_straylight = self.qc_straylight
        retv.region_name = dest_area
        retv.area = region
        retv.projection_name = region.proj_id

        retv.pcs_def = pcs_def_from_region(region)

        retv.num_of_columns = region.x_size
        retv.num_of_lines = region.y_size
        retv.xscale = region.pixel_size_x
        retv.yscale = region.pixel_size_y

        import pyproj
        prj = pyproj.Proj(region.proj4_string)
        aex = region.area_extent
        lonur, latur = prj(aex[2], aex[3], inverse=True)
        lonll, latll = prj(aex[0], aex[1], inverse=True)
        retv.ll_lon = lonll
        retv.ll_lat = latll
        retv.ur_lon = lonur
        retv.ur_lat = latur

        self.shape = region.shape

        retv.filled = True
        retv.resolution = self.resolution

        return retv

    def oldconvert2pps(self):
        """Converts the NWCSAF/MSG Cloud Type to the PPS format,
        in order to have consistency in output format between PPS and MSG.
        """
        retv = PpsCloudType()
        retv.region = SafRegion()
        retv.region.xsize = self.num_of_columns
        retv.region.ysize = self.num_of_lines
        retv.region.id = self.region_name
        retv.region.pcs_id = self.projection_name

        retv.region.pcs_def = pcs_def_from_region(self.area)
        retv.region.area_extent = self.area.area_extent
        retv.satellite_id = self.satid

        luts = pps_luts()
        retv.cloudtype_lut = luts[0]
        retv.phaseflag_lut = []
        retv.qualityflag_lut = []
        retv.cloudtype_des = "MSG SEVIRI Cloud Type"
        retv.qualityflag_des = 'MSG SEVIRI bitwise quality/processing flags'
        retv.phaseflag_des = 'MSG SEVIRI Cloud phase flags'

        retv.cloudtype = self.cloudtype.astype('B')
        retv.phaseflag = self.cloudphase.astype('B')
        retv.qualityflag = ctype_procflags2pps(self.processing_flags)

        return retv

    def convert2pps(self):
        from mpop.satin.nwcsaf_pps import CloudType
        from nwcsaf_formats.pps_conversions import (old_ctype_palette,
                                                    old_ctype_palette_data)

        retv = CloudType()

        region_type = np.dtype([('area_extent', '<f8', (4,)),
                                ('xsize', '<i4'),
                                ('ysize', '<i4'),
                                ('xscale', '<f4'),
                                ('yscale', '<f4'),
                                ('lat_0', '<f4'),
                                ('lon_0', '<f4'),
                                ('lat_ts', '<f4'),
                                ('id', 'S64'),
                                ('name', 'S64'),
                                ('pcs_id', 'S64'),
                                ('pcs_def', 'S128')])

        region = np.zeros((1, ), dtype=region_type)

        region["xsize"] = self.num_of_columns
        region["ysize"] = self.num_of_lines
        region["id"] = self.region_name
        region["pcs_id"] = self.projection_name

        region["pcs_def"] = pcs_def_from_region(self.area)
        region["area_extent"] = self.area.area_extent

        retv.region = InfoObject()
        retv.region.data = region
        retv._keys.append("region")

        retv.Region = region_type
        retv._keys.append("Region")

        retv._md["time_slot"] = self.nominal_product_time
        retv._md["description"] = np.string_(
            "MSG SEVIRI Cloud Type")
        retv._md["satellite_id"] = np.string_(self.satid)
        retv._md["sec_1970"] = np.uint64(0)
        retv._md["version"] = np.string_(self.product_algorithm_version)
        retv._md["orbit_number"] = np.uint64(0)

        retv.PALETTE = InfoObject()
        retv.PALETTE.data = old_ctype_palette_data()
        retv.PALETTE.info["CLASS"] = np.string_("PALETTE\0")
        retv.PALETTE.info["PAL_COLORMODEL"] = np.string_("RGB\0")
        retv.PALETTE.info["PAL_TYPE"] = np.string_("STANDARD8\0")
        retv.PALETTE.info["PAL_VERSION"] = np.string_("1.2\0")
        retv._keys.append("PALETTE")

        namelist = np.dtype([('outval_name', 'S128')])
        retv.OutputValueNameList = namelist
        retv._keys.append("OutputValueNameList")

        retv.cloudtype = InfoObject()
        retv.cloudtype.info["output_value_namelist"] = np.array(ctype_lut,
                                                                dtype=namelist)
        retv.cloudtype.info["CLASS"] = np.string_("IMAGE\0")
        retv.cloudtype.info["IMAGE_VERSION"] = np.string_("1.2\0")
        retv._refs[("cloudtype", "PALETTE")] = np.string_("PALETTE\0")
        retv.cloudtype.info["description"] = np.string_(
            "MSG SEVIRI Cloud Type")
        retv.cloudtype.data = self.cloudtype.astype('B')
        retv._projectables.append("cloudtype")

        # retv.PHASE_PALETTE = InfoObject()
        # retv.PHASE_PALETTE.data = self.cloudphase_palette
        # retv.PHASE_PALETTE.info["CLASS"] = np.string_("PALETTE")
        # retv.PHASE_PALETTE.info["PAL_COLORMODEL"] = np.string_("RGB")
        # retv.PHASE_PALETTE.info["PAL_TYPE"] = np.string_("STANDARD8")
        # retv.PHASE_PALETTE.info["PAL_VERSION"] = np.string_("1.2")
        # retv._keys.append("PHASE_PALETTE")

        retv.phase_flag = InfoObject()
        retv.phase_flag.info["output_value_namelist"] = np.array(phase_lut,
                                                                 dtype=namelist)
        # retv.phase_flag.info["CLASS"] = np.string_("IMAGE")
        # retv.phase_flag.info["IMAGE_VERSION"] = np.string_("1.2")
        # retv._refs[("phase_flag", "PALETTE")] = np.string_("PHASE_PALETTE")
        retv.phase_flag.info["description"] = np.string_(
            'MSG SEVIRI Cloud phase flags')
        retv.phase_flag.data = self.cloudphase.astype('B')
        retv._projectables.append("phase_flag")

        retv.quality_flag = InfoObject()
        retv.quality_flag.info["output_value_namelist"] = np.array(quality_lut,
                                                                   dtype=namelist)
        retv.quality_flag.info[
            "description"] = np.string_('MSG SEVIRI bitwise quality/processing flags')
        retv.quality_flag.data = ctype_procflags2pps(self.processing_flags)
        retv._projectables.append("quality_flag")

        retv.save = retv.write
        return retv

    def convert2nordrad(self):
        return NordRadCType(self)


class MsgCTTHData(object):

    """CTTH data object.
    """

    def __init__(self):
        self.data = None
        self.scaling_factor = 1
        self.offset = 0
        self.num_of_lines = 0
        self.num_of_columns = 0
        self.product = ""
        self.id = ""


class MsgCTTH(mpop.channel.GenericChannel):

    """CTTH channel.
    """

    def __init__(self, resolution=None):
        mpop.channel.GenericChannel.__init__(self, "CTTH")
        self.filled = False
        self.name = "CTTH"
        self.resolution = resolution
        self.package = ""
        self.saf = ""
        self.product_name = ""
        self.num_of_columns = 0
        self.num_of_lines = 0
        self.projection_name = ""
        self.region_name = ""
        self.cfac = 0
        self.lfac = 0
        self.coff = 0
        self.loff = 0
        self.nb_param = 0
        self.gp_sc_id = 0
        self.image_acquisition_time = 0
        self.spectral_channel_id = 0
        self.nominal_product_time = 0
        self.sgs_product_quality = 0
        self.sgs_product_completeness = 0
        self.product_algorithm_version = ""
        self.cloudiness = None  # Effective cloudiness
        self.processing_flags = None
        self.height = None
        self.temperature = None
        self.pressure = None
        self.satid = ""

    def __str__(self):
        return ("'%s: shape %s, resolution %sm'" %
                (self.name,
                 self.shape,
                 self.resolution))

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return self.filled

    def read(self, filename, calibrate=True):
        import h5py

        self.cloudiness = MsgCTTHData()  # Effective cloudiness
        self.temperature = MsgCTTHData()
        self.height = MsgCTTHData()
        self.pressure = MsgCTTHData()
        self.processing_flags = MsgCTTHData()

        h5f = h5py.File(filename, 'r')

        # The header
        # pylint: disable-msg=W0212
        self.package = h5f.attrs["PACKAGE"]
        self.saf = h5f.attrs["SAF"]
        self.product_name = h5f.attrs["PRODUCT_NAME"]
        self.num_of_columns = h5f.attrs["NC"]
        self.num_of_lines = h5f.attrs["NL"]
        self.projection_name = h5f.attrs["PROJECTION_NAME"]
        self.region_name = h5f.attrs["REGION_NAME"]
        self.cfac = h5f.attrs["CFAC"]
        self.lfac = h5f.attrs["LFAC"]
        self.coff = h5f.attrs["COFF"]
        self.loff = h5f.attrs["LOFF"]
        self.nb_param = h5f.attrs["NB_PARAMETERS"]
        self.gp_sc_id = h5f.attrs["GP_SC_ID"]
        self.image_acquisition_time = h5f.attrs["IMAGE_ACQUISITION_TIME"]
        self.spectral_channel_id = h5f.attrs["SPECTRAL_CHANNEL_ID"]
        self.nominal_product_time = datetime.strptime(h5f.attrs["NOMINAL_PRODUCT_TIME"],
                                                      "%Y%m%d%H%M")
        self.sgs_product_quality = h5f.attrs["SGS_PRODUCT_QUALITY"]
        self.sgs_product_completeness = h5f.attrs["SGS_PRODUCT_COMPLETENESS"]
        self.product_algorithm_version = h5f.attrs[
            "PACKAGE"] + h5f.attrs["PRODUCT_NAME"] + h5f.attrs["PRODUCT_ALGORITHM_VERSION"]
        # pylint: enable-msg=W0212
        # ------------------------

        # The CTTH cloudiness data
        if 'CTTH_EFFECT' not in h5f.keys():
            raise IOError("No field CTTH_EFFECT in product " +
                          str(self.product_name) +
                          "\n\tPerhaps you have found the wrong file for this product?")
        h5d = h5f['CTTH_EFFECT']
        self.cloudiness.data = h5d[:, :]
        self.cloudiness.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.cloudiness.offset = h5d.attrs["OFFSET"]
        self.cloudiness.num_of_lines = h5d.attrs["N_LINES"]
        self.cloudiness.num_of_columns = h5d.attrs["N_COLS"]
        self.cloudiness.product = h5d.attrs["PRODUCT"]
        self.cloudiness.id = h5d.attrs["ID"]

        self.cloudiness.data = np.ma.masked_equal(self.cloudiness.data, 255)
        self.cloudiness = np.ma.masked_equal(self.cloudiness.data, 0)
        self.cloudiness_palette = _get_palette(h5f, 'CTTH_EFFECT')

        # ------------------------

        # The CTTH temperature data
        h5d = h5f['CTTH_TEMPER']
        self.temperature.data = h5d[:, :]
        self.temperature.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.temperature.offset = h5d.attrs["OFFSET"]
        self.temperature.num_of_lines = h5d.attrs["N_LINES"]
        self.shape = (self.temperature.num_of_lines,
                      self.temperature.num_of_columns)
        self.temperature.num_of_columns = h5d.attrs["N_COLS"]
        self.temperature.product = h5d.attrs["PRODUCT"]
        self.temperature.id = h5d.attrs["ID"]

        self.temperature.data = np.ma.masked_equal(self.temperature.data, 0)
        if calibrate:
            self.temperature = (self.temperature.data *
                                self.temperature.scaling_factor +
                                self.temperature.offset)
        else:
            self.temperature = self.temperature.data
        self.temperature_palette = _get_palette(h5f, 'CTTH_TEMPER')

        # ------------------------

        # The CTTH pressure data
        h5d = h5f['CTTH_PRESS']
        self.pressure.data = h5d[:, :]
        self.pressure.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.pressure.offset = h5d.attrs["OFFSET"]
        self.pressure.num_of_lines = h5d.attrs["N_LINES"]
        self.pressure.num_of_columns = h5d.attrs["N_COLS"]
        self.pressure.product = h5d.attrs["PRODUCT"]
        self.pressure.id = h5d.attrs["ID"]

        self.pressure.data = np.ma.masked_equal(self.pressure.data, 255)
        self.pressure.data = np.ma.masked_equal(self.pressure.data, 0)
        if calibrate:
            self.pressure = (self.pressure.data *
                             self.pressure.scaling_factor +
                             self.pressure.offset)
        else:
            self.pressure = self.pressure.data
        self.pressure_palette = _get_palette(h5f, 'CTTH_PRESS')

        # ------------------------

        # The CTTH height data
        h5d = h5f['CTTH_HEIGHT']
        self.height.data = h5d[:, :]
        self.height.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.height.offset = h5d.attrs["OFFSET"]
        self.height.num_of_lines = h5d.attrs["N_LINES"]
        self.height.num_of_columns = h5d.attrs["N_COLS"]
        self.height.product = h5d.attrs["PRODUCT"]
        self.height.id = h5d.attrs["ID"]

        self.height.data = np.ma.masked_equal(self.height.data, 255)
        self.height.data = np.ma.masked_equal(self.height.data, 0)
        if calibrate:
            self.height = (self.height.data *
                           self.height.scaling_factor +
                           self.height.offset)
        else:
            self.height = self.height.data
        self.height_palette = _get_palette(h5f, 'CTTH_HEIGHT')

        # ------------------------

        # The CTTH processing/quality flags
        h5d = h5f['CTTH_QUALITY']
        self.processing_flags.data = h5d[:, :]
        self.processing_flags.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.processing_flags.offset = h5d.attrs["OFFSET"]
        self.processing_flags.num_of_lines = \
            h5d.attrs["N_LINES"]
        self.processing_flags.num_of_columns = \
            h5d.attrs["N_COLS"]
        self.processing_flags.product = h5d.attrs["PRODUCT"]
        self.processing_flags.id = h5d.attrs["ID"]

        self.processing_flags = \
            np.ma.masked_equal(self.processing_flags.data, 0)

        h5f.close()

        self.shape = self.height.shape

        self.area = get_area_from_file(filename)

        self.filled = True

    def save(self, filename, **kwargs):
        """Save the current CTTH channel to HDF5 format.
        """
        ctth = self.convert2pps()
        LOG.info("Saving CTTH hdf file...")
        ctth.save(filename, **kwargs)
        LOG.info("Saving CTTH hdf file done !")

    def project(self, coverage):
        """Project the current CTTH channel along the *coverage*
        """
        dest_area = coverage.out_area
        dest_area_id = dest_area.area_id

        retv = MsgCTTH()

        retv.temperature = coverage.project_array(self.temperature)
        retv.height = coverage.project_array(self.height)
        retv.pressure = coverage.project_array(self.pressure)
        retv.cloudiness = coverage.project_array(self.cloudiness)
        retv.processing_flags = \
            coverage.project_array(self.processing_flags)

        retv.area = dest_area
        retv.region_name = dest_area_id
        retv.projection_name = dest_area.proj_id
        retv.num_of_columns = dest_area.x_size
        retv.num_of_lines = dest_area.y_size

        retv.shape = dest_area.shape

        retv.name = self.name
        retv.resolution = self.resolution
        retv.filled = True

        return retv

# ------------------------------------------------------------------
    def oldconvert2pps(self):
        """Convert the current CTTH channel to pps format.
        """
        retv = PpsCTTH()
        retv.region = SafRegion()
        retv.region.xsize = self.num_of_columns
        retv.region.ysize = self.num_of_lines
        retv.region.id = self.region_name
        retv.region.pcs_id = self.projection_name
        retv.region.pcs_def = pcs_def_from_region(self.area)
        retv.region.area_extent = self.area.area_extent
        retv.satellite_id = self.satid

        retv.processingflag_lut = []
        retv.des = "MSG SEVIRI Cloud Top Temperature & Height"
        retv.ctt_des = "MSG SEVIRI cloud top temperature (K)"
        retv.ctp_des = "MSG SEVIRI cloud top pressure (hPa)"
        retv.ctp_des = "MSG SEVIRI cloud top height (m)"
        retv.cloudiness_des = "MSG SEVIRI effective cloudiness (%)"
        retv.processingflag_des = 'MSG SEVIRI bitwise quality/processing flags'

        retv.t_gain = 1.0
        retv.t_intercept = 100.0
        retv.t_nodata = 255

        retv.temperature = ((self.temperature - retv.t_intercept) /
                            retv.t_gain).filled(retv.t_nodata).astype('B')

        retv.h_gain = 200.0
        retv.h_intercept = 0.0
        retv.h_nodata = 255

        retv.height = ((self.height - retv.h_intercept) /
                       retv.h_gain).filled(retv.h_nodata).astype('B')

        retv.p_gain = 25.0
        retv.p_intercept = 0.0
        retv.p_nodata = 255

        retv.pressure = ((self.pressure - retv.p_intercept) /
                         retv.p_gain).filled(retv.p_nodata).astype('B')

        retv.cloudiness = self.cloudiness.astype('B')
        retv.c_nodata = 255  # Is this correct? FIXME

        retv.processingflag = ctth_procflags2pps(self.processing_flags)

        return retv

    def convert2pps(self):
        """Convert the current CTTH channel to pps format.
        """
        from mpop.satin.nwcsaf_pps import CloudTopTemperatureHeight
        from nwcsaf_formats.pps_conversions import (old_ctth_press_palette_data,
                                                    old_ctth_temp_palette_data,
                                                    old_ctth_height_palette_data)

        retv = CloudTopTemperatureHeight()

        region_type = np.dtype([('area_extent', '<f8', (4,)),
                                ('xsize', '<i4'),
                                ('ysize', '<i4'),
                                ('xscale', '<f4'),
                                ('yscale', '<f4'),
                                ('lat_0', '<f4'),
                                ('lon_0', '<f4'),
                                ('lat_ts', '<f4'),
                                ('id', 'S64'),
                                ('name', 'S64'),
                                ('pcs_id', 'S128'),
                                ('pcs_def', 'S128')])

        region = np.zeros((1, ), dtype=region_type)

        region["xsize"] = self.num_of_columns
        region["ysize"] = self.num_of_lines
        region["id"] = self.region_name
        region["pcs_id"] = self.projection_name

        region["pcs_def"] = pcs_def_from_region(self.area)
        region["area_extent"] = self.area.area_extent

        retv.Region = region_type
        retv._keys.append("Region")

        namelist = np.dtype([('outval_name', 'S128')])
        retv.OutputValueNameList = namelist
        retv._keys.append("OutputValueNameList")

        retv.region = InfoObject()
        retv.region.data = region
        retv._keys.append("region")

        retv._md["time_slot"] = self.nominal_product_time
        retv._md["description"] = np.string_(
            "MSG SEVIRI Cloud Top Temperature & Height")
        retv._md["satellite_id"] = np.string_(self.satid)
        retv._md["sec_1970"] = np.uint64(0)
        retv._md["version"] = np.string_(self.product_algorithm_version)
        retv._md["orbit_number"] = np.uint64(0)
        retv._md["version"] = np.string_(self.product_algorithm_version)

        retv.processingflag_lut = []

        retv.cloudiness = InfoObject()
        retv.cloudiness.info["description"] = np.string_(
            "MSG SEVIRI effective cloudiness (%)")
        retv.cloudiness.info["gain"] = np.float32(0.0)
        retv.cloudiness.info["intercept"] = np.float32(0.0)
        retv.cloudiness.info["no_data_value"] = np.uint8(255)
        retv.cloudiness.data = self.cloudiness.astype('B')
        retv._projectables.append("cloudiness")

        retv.HEIGHT_PALETTE = InfoObject()
        retv.HEIGHT_PALETTE.data = old_ctth_height_palette_data()
        retv.HEIGHT_PALETTE.info["CLASS"] = np.string_("PALETTE")
        retv.HEIGHT_PALETTE.info["PAL_COLORMODEL"] = np.string_("RGB")
        retv.HEIGHT_PALETTE.info["PAL_TYPE"] = np.string_("STANDARD8")
        retv.HEIGHT_PALETTE.info["PAL_VERSION"] = np.string_("1.2")
        retv._keys.append("HEIGHT_PALETTE")

        retv.height = InfoObject()
        retv.height.info["description"] = np.string_(
            "MSG SEVIRI cloud top height (m)")
        retv.height.info["gain"] = np.float32(200.0)
        retv.height.info["intercept"] = np.float32(0.0)
        retv.height.info["no_data_value"] = np.uint8(255)
        retv.height.data = ((self.height - 0.0) /
                            200.0).filled(255).astype('B')
        retv.height.info["CLASS"] = np.string_("IMAGE")
        retv.height.info["IMAGE_VERSION"] = np.string_("1.2")
        retv._refs[("height", "PALETTE")] = np.string_("HEIGHT_PALETTE")
        retv._projectables.append("height")

        retv.PRESSURE_PALETTE = InfoObject()
        retv.PRESSURE_PALETTE.data = old_ctth_press_palette_data()
        retv.PRESSURE_PALETTE.info["CLASS"] = np.string_("PALETTE")
        retv.PRESSURE_PALETTE.info["PAL_COLORMODEL"] = np.string_("RGB")
        retv.PRESSURE_PALETTE.info["PAL_TYPE"] = np.string_("STANDARD8")
        retv.PRESSURE_PALETTE.info["PAL_VERSION"] = np.string_("1.2")
        retv._keys.append("PRESSURE_PALETTE")

        retv.pressure = InfoObject()
        retv.pressure.info["description"] = \
            np.string_("MSG SEVIRI cloud top pressure (hPa)")
        retv.pressure.info["gain"] = np.float32(25.0)
        retv.pressure.info["intercept"] = np.float32(0.0)
        retv.pressure.info["no_data_value"] = np.uint8(255)
        retv.pressure.data = ((self.pressure - 0.0) /
                              25.0).filled(255).astype('B')
        retv.pressure.info["CLASS"] = np.string_("IMAGE")
        retv.pressure.info["IMAGE_VERSION"] = np.string_("1.2")
        retv._refs[("pressure", "PALETTE")] = np.string_("PRESSURE_PALETTE")
        retv._projectables.append("pressure")

        retv.TEMPERATURE_PALETTE = InfoObject()
        retv.TEMPERATURE_PALETTE.data = old_ctth_temp_palette_data()
        retv.TEMPERATURE_PALETTE.info["CLASS"] = np.string_("PALETTE")
        retv.TEMPERATURE_PALETTE.info["PAL_COLORMODEL"] = np.string_("RGB")
        retv.TEMPERATURE_PALETTE.info["PAL_TYPE"] = np.string_("STANDARD8")
        retv.TEMPERATURE_PALETTE.info["PAL_VERSION"] = np.string_("1.2")
        retv._keys.append("TEMPERATURE_PALETTE")

        retv.temperature = InfoObject()
        retv.temperature.info["description"] = \
            np.string_("MSG SEVIRI cloud top temperature (K)")
        retv.temperature.info["gain"] = np.float32(1.0)
        retv.temperature.info["intercept"] = np.float32(100.0)
        retv.temperature.info["no_data_value"] = np.uint8(255)
        retv.temperature.data = ((self.temperature - 100.0) /
                                 1.0).filled(255).astype('B')
        retv.temperature.info["CLASS"] = np.string_("IMAGE")
        retv.temperature.info["IMAGE_VERSION"] = np.string_("1.2")
        retv._refs[("temperature", "PALETTE")] = np.string_(
            "TEMPERATURE_PALETTE")
        retv._projectables.append("temperature")

        retv.processing_flag = InfoObject()
        # retv.processing_flag.info["output_value_nameslist"] = processing_lut
        retv.processing_flag.info["output_value_namelist"] = np.array(processing_flags_lut,
                                                                      dtype=namelist)
        retv.processing_flag.info["description"] = np.string_(
            'MSG SEVIRI bitwise quality/processing flags')
        retv.processing_flag.data = ctth_procflags2pps(self.processing_flags)
        retv._projectables.append("processing_flag")

        retv.save = retv.write

        return retv

# ----------------------------------------


class MsgPCData(object):

    """NWCSAF/MSG Precipitating Clouds data layer
    """

    def __init__(self):
        self.data = None
        self.scaling_factor = 1
        self.offset = 0
        self.num_of_lines = 0
        self.num_of_columns = 0
        self.product = ""
        self.id = ""


class MsgPC(mpop.channel.GenericChannel):

    """NWCSAF/MSG Precipitating Clouds data structure as retrieved from HDF5
    file. Resolution sets the nominal resolution of the data.
    """

    def __init__(self):
        mpop.channel.GenericChannel.__init__(self, "PC")
        self.filled = False
        self.name = "PC"
        self.package = ""
        self.saf = ""
        self.product_name = ""
        self.num_of_columns = 0
        self.num_of_lines = 0
        self.projection_name = ""
        self.pcs_def = ""
        self.xscale = 0
        self.yscale = 0
        self.ll_lon = 0.0
        self.ll_lat = 0.0
        self.ur_lon = 0.0
        self.ur_lat = 0.0
        self.region_name = ""
        self.cfac = 0
        self.lfac = 0
        self.coff = 0
        self.loff = 0
        self.nb_param = 0
        self.gp_sc_id = 0
        self.image_acquisition_time = 0
        self.spectral_channel_id = 0
        self.nominal_product_time = 0
        self.sgs_product_quality = 0
        self.sgs_product_completeness = 0
        self.product_algorithm_version = ""
        self.probability_1 = None
        self.processing_flags = None
        self.shape = None
        self.satid = ""
        self.qc_straylight = -1

    def __str__(self):
        return ("'%s: shape %s, resolution %sm'" %
                (self.name,
                 self.probability_1.shape,
                 self.resolution))

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return self.filled

# ------------------------------------------------------------------
    def read(self, filename, calibrate=True):
        """Reader for the NWCSAF/MSG precipitating clouds. Use *filename* to read data.
        """
        import h5py

        self.probability_1 = MsgPCData()
        self.processing_flags = MsgPCData()

        h5f = h5py.File(filename, 'r')
        # pylint: disable-msg=W0212
        self.package = h5f.attrs["PACKAGE"]
        self.saf = h5f.attrs["SAF"]
        self.product_name = h5f.attrs["PRODUCT_NAME"]
        self.num_of_columns = h5f.attrs["NC"]
        self.num_of_lines = h5f.attrs["NL"]
        self.projection_name = h5f.attrs["PROJECTION_NAME"]
        self.region_name = h5f.attrs["REGION_NAME"]
        self.cfac = h5f.attrs["CFAC"]
        self.lfac = h5f.attrs["LFAC"]
        self.coff = h5f.attrs["COFF"]
        self.loff = h5f.attrs["LOFF"]
        self.nb_param = h5f.attrs["NB_PARAMETERS"]
        self.gp_sc_id = h5f.attrs["GP_SC_ID"]
        self.image_acquisition_time = h5f.attrs["IMAGE_ACQUISITION_TIME"]
        self.spectral_channel_id = h5f.attrs["SPECTRAL_CHANNEL_ID"]
        self.nominal_product_time = datetime.strptime(h5f.attrs["NOMINAL_PRODUCT_TIME"],
                                                      "%Y%m%d%H%M")
        self.sgs_product_quality = h5f.attrs["SGS_PRODUCT_QUALITY"]
        self.sgs_product_completeness = h5f.attrs["SGS_PRODUCT_COMPLETENESS"]
        self.product_algorithm_version = h5f.attrs[
            "PACKAGE"] + h5f.attrs["PRODUCT_NAME"] + h5f.attrs["PRODUCT_ALGORITHM_VERSION"]

        # pylint: enable-msg=W0212
        # ------------------------

        # The precipitating clouds data
        h5d = h5f['PC_PROB1']
        self.probability_1.data = h5d[:, :]
        self.probability_1.scaling_factor = h5d.attrs["SCALING_FACTOR"]
        self.probability_1.offset = h5d.attrs["OFFSET"]
        self.probability_1.num_of_lines = h5d.attrs["N_LINES"]
        self.probability_1.num_of_columns = h5d.attrs["N_COLS"]
        self.shape = (self.probability_1.num_of_lines,
                      self.probability_1.num_of_columns)
        self.probability_1.product = h5d.attrs["PRODUCT"]
        self.probability_1.id = h5d.attrs["ID"]
        self.probability_1.data = np.ma.masked_equal(
            self.probability_1.data, 0)
        if calibrate:
            self.probability_1 = (self.probability_1.data *
                                  self.probability_1.scaling_factor +
                                  self.probability_1.offset)
        else:
            self.probability_1 = self.probability_1.data
        self.probability_1_palette = _get_palette(h5f, 'PC_PROB1')

        # ------------------------

        # The cloudtype processing/quality flags
        h5d = h5f['PC_QUALITY']
        self.processing_flags.data = h5d[:, :]
        self.processing_flags.scaling_factor = \
            h5d.attrs["SCALING_FACTOR"]
        self.processing_flags.offset = h5d.attrs["OFFSET"]
        self.processing_flags.num_of_lines = h5d.attrs["N_LINES"]
        self.processing_flags.num_of_columns = h5d.attrs["N_COLS"]
        self.processing_flags.product = h5d.attrs["PRODUCT"]
        self.processing_flags.id = h5d.attrs["ID"]
        self.processing_flags = np.ma.masked_equal(
            self.processing_flags.data, 0)

        # ------------------------
        h5f.close()

        self.area = get_area_from_file(filename)

        self.filled = True

# ------------------------------------------------------------------


def get_bit_from_flags(arr, nbit):
    """I don't know what this function does.
    """
    res = np.bitwise_and(np.right_shift(arr, nbit), 1)
    return res.astype('b')


def ctth_procflags2pps(data):
    """Convert ctth processing flags from MSG to PPS format.
    """

    ones = np.ones(data.shape, "h")

    # 2 bits to define processing status
    # (maps to pps bits 0 and 1:)
    is_bit0_set = get_bit_from_flags(data, 0)
    is_bit1_set = get_bit_from_flags(data, 1)
    proc = (is_bit0_set * np.left_shift(ones, 0) +
            is_bit1_set * np.left_shift(ones, 1))
    del is_bit0_set
    del is_bit1_set

    # Non-processed?
    # If non-processed in msg (0) then set pps bit 0 and nothing else.
    # If non-processed in msg due to FOV is cloud free (1) then do not set any
    # pps bits.
    # If processed (because cloudy) with/without result in msg (2&3) then set
    # pps bit 1.

    arr = np.where(np.equal(proc, 0), np.left_shift(ones, 0), 0)
    arr = np.where(np.equal(proc, 2), np.left_shift(ones, 1), 0)
    arr = np.where(np.equal(proc, 3), np.left_shift(ones, 1), 0)
    retv = np.array(arr)
    del proc

    # 1 bit to define if RTTOV-simulations are available?
    # (maps to pps bit 3:)
    is_bit2_set = get_bit_from_flags(data, 2)
    proc = is_bit2_set

    # RTTOV-simulations available?

    arr = np.where(np.equal(proc, 1), np.left_shift(ones, 3), 0)
    retv = np.add(retv, arr)
    del is_bit2_set

    # 3 bits to describe NWP input data
    # (maps to pps bits 4&5:)
    is_bit3_set = get_bit_from_flags(data, 3)
    is_bit4_set = get_bit_from_flags(data, 4)
    is_bit5_set = get_bit_from_flags(data, 5)
    # Put together the three bits into a nwp-flag:
    nwp_bits = (is_bit3_set * np.left_shift(ones, 0) +
                is_bit4_set * np.left_shift(ones, 1) +
                is_bit5_set * np.left_shift(ones, 2))
    arr = np.where(np.logical_and(np.greater_equal(nwp_bits, 3),
                                  np.less_equal(nwp_bits, 5)),
                   np.left_shift(ones, 4),
                   0)
    arr = np.add(arr, np.where(np.logical_or(np.equal(nwp_bits, 2),
                                             np.equal(nwp_bits, 4)),
                               np.left_shift(ones, 5),
                               0))

    retv = np.add(retv, arr)
    del is_bit3_set
    del is_bit4_set
    del is_bit5_set

    # 2 bits to describe SEVIRI input data
    # (maps to pps bits 6:)
    is_bit6_set = get_bit_from_flags(data, 6)
    is_bit7_set = get_bit_from_flags(data, 7)
    # Put together the two bits into a seviri-flag:
    seviri_bits = (is_bit6_set * np.left_shift(ones, 0) +
                   is_bit7_set * np.left_shift(ones, 1))
    arr = np.where(np.greater_equal(seviri_bits, 2),
                   np.left_shift(ones, 6), 0)

    retv = np.add(retv, arr)
    del is_bit6_set
    del is_bit7_set

    # 4 bits to describe which method has been used
    # (maps to pps bits 7&8 and bit 2:)
    is_bit8_set = get_bit_from_flags(data, 8)
    is_bit9_set = get_bit_from_flags(data, 9)
    is_bit10_set = get_bit_from_flags(data, 10)
    is_bit11_set = get_bit_from_flags(data, 11)
    # Put together the four bits into a method-flag:
    method_bits = (is_bit8_set * np.left_shift(ones, 0) +
                   is_bit9_set * np.left_shift(ones, 1) +
                   is_bit10_set * np.left_shift(ones, 2) +
                   is_bit11_set * np.left_shift(ones, 3))
    arr = np.where(np.logical_or(
        np.logical_and(np.greater_equal(method_bits, 1),
                       np.less_equal(method_bits, 2)),
        np.equal(method_bits, 13)),
        np.left_shift(ones, 2),
        0)
    arr = np.add(arr,
                 np.where(np.equal(method_bits, 1),
                          np.left_shift(ones, 7),
                          0))
    arr = np.add(arr,
                 np.where(np.logical_and(
                     np.greater_equal(method_bits, 3),
                     np.less_equal(method_bits, 12)),
                     np.left_shift(ones, 8),
                     0))

    # (Maps directly - as well - to the spare bits 9-12)
    arr = np.add(arr, np.where(is_bit8_set, np.left_shift(ones, 9), 0))
    arr = np.add(arr, np.where(is_bit9_set,
                               np.left_shift(ones, 10),
                               0))
    arr = np.add(arr, np.where(is_bit10_set,
                               np.left_shift(ones, 11),
                               0))
    arr = np.add(arr, np.where(is_bit11_set,
                               np.left_shift(ones, 12),
                               0))
    retv = np.add(retv, arr)
    del is_bit8_set
    del is_bit9_set
    del is_bit10_set
    del is_bit11_set

    # 2 bits to describe the quality of the processing itself
    # (maps to pps bits 14&15:)
    is_bit12_set = get_bit_from_flags(data, 12)
    is_bit13_set = get_bit_from_flags(data, 13)
    # Put together the two bits into a quality-flag:
    qual_bits = (is_bit12_set * np.left_shift(ones, 0) +
                 is_bit13_set * np.left_shift(ones, 1))
    arr = np.where(np.logical_and(np.greater_equal(qual_bits, 1),
                                  np.less_equal(qual_bits, 2)),
                   np.left_shift(ones, 14), 0)
    arr = np.add(arr,
                 np.where(np.equal(qual_bits, 2),
                          np.left_shift(ones, 15),
                          0))

    retv = np.add(retv, arr)
    del is_bit12_set
    del is_bit13_set

    return retv.astype('h')


def ctype_procflags2pps(data):
    """Converting cloud type processing flags to
    the PPS format, in order to have consistency between
    PPS and MSG cloud type contents.
    """

    ones = np.ones(data.shape, "h")

    # msg illumination bit 0,1,2 (undefined,night,twilight,day,sunglint) maps
    # to pps bits 2, 3 and 4:
    is_bit0_set = get_bit_from_flags(data, 0)
    is_bit1_set = get_bit_from_flags(data, 1)
    is_bit2_set = get_bit_from_flags(data, 2)
    illum = is_bit0_set * np.left_shift(ones, 0) + \
        is_bit1_set * np.left_shift(ones, 1) + \
        is_bit2_set * np.left_shift(ones, 2)
    del is_bit0_set
    del is_bit1_set
    del is_bit2_set
    # Night?
    # If night in msg then set pps night bit and nothing else.
    # If twilight in msg then set pps twilight bit and nothing else.
    # If day in msg then unset both the pps night and twilight bits.
    # If sunglint in msg unset both the pps night and twilight bits and set the
    # pps sunglint bit.
    arr = np.where(np.equal(illum, 1), np.left_shift(ones, 2), 0)
    arr = np.where(np.equal(illum, 2), np.left_shift(ones, 3), arr)
    arr = np.where(np.equal(illum, 3), 0, arr)
    arr = np.where(np.equal(illum, 4), np.left_shift(ones, 4), arr)
    retv = np.array(arr)
    del illum

    # msg nwp-input bit 3 (nwp present?) maps to pps bit 7:
    # msg nwp-input bit 4 (low level inversion?) maps to pps bit 6:
    is_bit3_set = get_bit_from_flags(data, 3)
    is_bit4_set = get_bit_from_flags(data, 4)
    nwp = (is_bit3_set * np.left_shift(ones, 0) +
           is_bit4_set * np.left_shift(ones, 1))
    del is_bit3_set
    del is_bit4_set

    arr = np.where(np.equal(nwp, 1), np.left_shift(ones, 7), 0)
    arr = np.where(np.equal(nwp, 2), np.left_shift(ones, 7) +
                   np.left_shift(ones, 6), arr)
    arr = np.where(np.equal(nwp, 3), 0, arr)
    retv = np.add(arr, retv)
    del nwp

    # msg seviri-input bits 5&6 maps to pps bit 8:
    is_bit5_set = get_bit_from_flags(data, 5)
    is_bit6_set = get_bit_from_flags(data, 6)
    seviri = (is_bit5_set * np.left_shift(ones, 0) +
              is_bit6_set * np.left_shift(ones, 1))
    del is_bit5_set
    del is_bit6_set

    retv = np.add(retv,
                  np.where(np.logical_or(np.equal(seviri, 2),
                                         np.equal(seviri, 3)),
                           np.left_shift(ones, 8), 0))
    del seviri

    # msg quality bits 7&8 maps to pps bit 9&10:
    is_bit7_set = get_bit_from_flags(data, 7)
    is_bit8_set = get_bit_from_flags(data, 8)
    quality = (is_bit7_set * np.left_shift(ones, 0) +
               is_bit8_set * np.left_shift(ones, 1))
    del is_bit7_set
    del is_bit8_set

    arr = np.where(np.equal(quality, 2), np.left_shift(ones, 9), 0)
    arr = np.where(np.equal(quality, 3), np.left_shift(ones, 10), arr)
    retv = np.add(arr, retv)
    del quality

    # msg bit 9 (stratiform-cumuliform distinction?) maps to pps bit 11:
    is_bit9_set = get_bit_from_flags(data, 9)
    retv = np.add(retv,
                  np.where(is_bit9_set,
                           np.left_shift(ones, 11),
                           0))
    del is_bit9_set

    return retv.astype('h')


# def pps_luts():
#     """Gets the LUTs for the PPS Cloud Type data fields.
#     Returns a tuple with Cloud Type lut, Cloud Phase lut, Processing flags lut
#     """

#     return ctype_lut, phase_lut, quality_lut


def add_nullterm_attr(obj, key, val):
    import h5py
    atype = h5py.h5t.C_S1
    tid = h5py.h5t.TypeID.copy(h5py.h5t.C_S1)
    tid.set_size(len(val) + 1)
    tid.set_strpad(h5py.h5t.STR_NULLTERM)
    sid = h5py.h5s.create(h5py.h5s.SCALAR)
    aid = h5py.h5a.create(obj.id, key, tid, sid)
    blob = np.string_(val)
    aid.write(np.array(val))


class NordRadCType(object):

    """Wrapper aroud the msg_ctype channel.
    """

    def __init__(self, ctype_instance):
        self.ctype = ctype_instance
        self.datestr = ctype_instance.image_acquisition_time
        self.info = {}

    def save(self, filename, **kwargs):
        """Save the current instance to nordrad hdf format.
        """
        #import _pyhl
        import h5py

        msgctype = self.ctype

        h5f = h5py.File(filename, mode="w")

        # What
        what = h5f.create_group("what")
        #what.attrs["object"] = np.string_("IMAGE")
        add_nullterm_attr(what, "object", "IMAGE")
        what.attrs["sets"] = np.int(1)
        #what.attrs["version"] = np.string_("H5rad 1.2")
        add_nullterm_attr(what, "version", "H5rad 1.2")

        yyyymmdd = self.datestr[0:8]
        hourminsec = self.datestr[8:12] + '00'
        #what.attrs["date"] = np.string_(yyyymmdd)
        add_nullterm_attr(what, "date", yyyymmdd)
        #what.attrs["time"] = np.string_(hourminsec)
        add_nullterm_attr(what, "time", hourminsec)

        # Where
        where = h5f.create_group("where")
        #where.attrs["projdef"] = np.string_(msgctype.area.proj4_string)
        add_nullterm_attr(where, "projdef", msgctype.area.proj4_string)
        where.attrs["xsize"] = np.int(msgctype.num_of_columns)
        where.attrs["ysize"] = np.int(msgctype.num_of_lines)
        where.attrs["xscale"] = np.float(msgctype.xscale)
        where.attrs["yscale"] = np.float(msgctype.yscale)
        where.attrs["LL_lon"] = np.float(msgctype.ll_lon)
        where.attrs["LL_lat"] = np.float(msgctype.ll_lat)
        where.attrs["UR_lon"] = np.float(msgctype.ur_lon)
        where.attrs["UR_lat"] = np.float(msgctype.ur_lat)

        # How
        how = h5f.create_group("how")
        #how.attrs["area"] = np.string_(msgctype.region_name)
        add_nullterm_attr(how, "area", msgctype.region_name)

        # image1
        image1 = h5f.create_group("image1")
        image1.create_dataset("data", data=msgctype.cloudtype.astype('B'),
                              compression="gzip", compression_opts=COMPRESS_LVL)

        what = image1.create_group("what")
        #what.attrs["product"] = np.string_('MSGCT')
        add_nullterm_attr(what, "product", 'MSGCT')
        what.attrs["prodpar"] = np.float(0.0)
        #what.attrs["quantity"] = np.string_("ct")
        add_nullterm_attr(what, "quantity", "ct")
        #what.attrs["startdate"] = np.string_(yyyymmdd)
        add_nullterm_attr(what, "startdate", yyyymmdd)
        #what.attrs["starttime"] = np.string_(hourminsec)
        add_nullterm_attr(what, "starttime", hourminsec)
        #what.attrs["enddate"] = np.string_(yyyymmdd)
        add_nullterm_attr(what, "enddate", yyyymmdd)
        #what.attrs["endtime"] = np.string_(hourminsec)
        add_nullterm_attr(what, "endtime", hourminsec)
        what.attrs["gain"] = np.float(1.0)
        what.attrs["offset"] = np.float(0.0)
        what.attrs["nodata"] = np.float(0.0)
        # What we call missingdata in PPS:
        what.attrs["undetect"] = np.float(20.0)

        h5f.close()


MSG_PGE_EXTENTIONS = ["PLAX.CTTH.0.h5", "PLAX.CLIM.0.h5", "h5"]


def get_best_product(filename, area_extent):
    """Get the best of the available products for the *filename* template.
    """

    basename = filename
    for ext in MSG_PGE_EXTENTIONS:
        if filename.endswith(ext):
            basename = filename[:-(len(ext) + 1)]
            break
    for ext in MSG_PGE_EXTENTIONS:
        if not basename.endswith(ext):
            match_str = basename + "." + ext
        else:
            match_str = basename
        LOG.debug("glob-string for filename: " + str(match_str))
        flist = glob.glob(match_str)
        if len(flist) == 0:
            LOG.warning("No matching .%s input MSG file."
                        % ext)
        else:
            # File found:
            if area_extent is None:
                LOG.warning("Didn't specify an area, taking " + flist[0])
                return flist[0]
            for fname in flist:
                aex = get_area_extent(fname)
                if np.all(np.max(np.abs(np.array(aex) -
                                        np.array(area_extent))) < 1000):
                    LOG.info("MSG file found: %s" % fname)
                    return fname
            LOG.info("Did not find any MSG file for specified area")


def get_best_products(filename, area_extent):
    """Get the best of the available products for the *filename* template.
    """

    basename = filename
    for ext in MSG_PGE_EXTENTIONS:
        if filename.endswith(ext):
            basename = filename[:-(len(ext) + 1)]
            break

    filenames = []

    for ext in MSG_PGE_EXTENTIONS:
        if not basename.endswith(ext):
            match_str = basename + "." + ext
        else:
            match_str = basename

        LOG.debug('Match string = ' + str(match_str))
        flist = glob.glob(match_str)
        if len(flist) == 0:
            LOG.warning("No matching .%s input MSG file."
                        % ext)
        else:
            # File found:
            if area_extent is None:
                LOG.warning("Didn't specify an area, taking " + flist[0])
                filenames.append(flist[0])
            else:
                found = False
                for fname in flist:
                    aex = get_area_extent(fname)
                    if np.all(np.max(np.abs(np.array(aex) -
                                            np.array(area_extent))) < 1000):
                        found = True
                        LOG.info("MSG file found: %s" % fname)
                        filenames.append(fname)
                    if not found:
                        LOG.info(
                            "Did not find any MSG file for specified area")
    LOG.debug("Sorted filenames: %s", str(sorted(filenames)))
    return sorted(filenames)


def get_area_from_file(filename):
    """Get the area from the h5 file.
    """
    from pyresample.geometry import AreaDefinition
    import h5py

    aex = get_area_extent(filename)
    h5f = h5py.File(filename, 'r')
    pname = h5f.attrs["PROJECTION_NAME"]
    proj = {}
    if pname.startswith("GEOS"):
        proj["proj"] = "geos"
        proj["a"] = "6378169.0"
        proj["b"] = "6356583.8"
        proj["h"] = "35785831.0"
        proj["lon_0"] = str(float(pname.split("<")[1][:-1]))
    else:
        raise NotImplementedError("Only geos projection supported yet.")
    area_def = AreaDefinition(h5f.attrs["REGION_NAME"],
                              h5f.attrs["REGION_NAME"],
                              pname,
                              proj,
                              int(h5f.attrs["NC"]),
                              int(h5f.attrs["NL"]),
                              aex)
    h5f.close()
    return area_def


def load(scene, **kwargs):
    """Load data into the *channels*. *Channels* is a list or a tuple
    containing channels we will load data into. If None, all channels are
    loaded.
    """

    area_extent = kwargs.get("area_extent")
    conf = ConfigParser.ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, scene.fullname + ".cfg"))

    if kwargs.get("filename") is not None:
        full_filename = kwargs["filename"]
        LOG.debug("File name determined from input: " + str(full_filename))
    else:
        directory = conf.get(scene.instrument_name + "-level3",
                             "dir",
                             raw=True)
        filename = conf.get(scene.instrument_name + "-level3", "filename",
                            raw=True)
        pathname = os.path.join(directory, filename)
        full_filename = None

    for chan in scene.channels_to_load:
        LOG.debug("Channel to load = " + str(chan))
        if chan == "CTTH":
            product = 'CTTH_'
            pname = 'CTTH'
            number = '03'
        elif chan == "CloudType":
            product = 'CT___'
            pname = 'CloudType'
            number = '02'
        elif chan == "CloudType_plax":
            product = 'CT___'
            pname = 'CloudType_plax'
            number = '02'
        else:
            LOG.info('Channel ' + str(chan) + ' not known by this reader!')
            continue

        if not full_filename:
            filename = (scene.time_slot.strftime(pathname)
                        % {"number": number,
                           "product": product})
        else:
            filename = full_filename

        ct_chan = CASES[pname]()
        if pname == 'CTTH':
            product = get_best_product(filename, area_extent)
        elif pname == 'CloudType':
            product = get_best_products(filename, area_extent)[-1]
        elif pname == 'CloudType_plax':
            product = get_best_products(filename, area_extent)[0]
            LOG.debug("Parallax corrected file: %s", product)

        if not product or not os.path.exists(product):
            raise IOError("Failed to locate file...")

        ct_chan.read(product)
        ct_chan.name = pname
        ct_chan.satid = (scene.fullname.capitalize())
        ct_chan.resolution = ct_chan.area.pixel_size_x
        scene.channels.append(ct_chan)

        LOG.info("Channel " + str(chan) + " loaded")

    LOG.info("Loading channels done.")


CASES = {
    'CTTH': MsgCTTH,
    'CloudType': MsgCloudType,
    'CloudType_plax': MsgCloudType
}

if __name__ == '__main__':

    FILENAME = "/data/proj/safutv/geo_out/0deg/SAFNWC_MSG3_CT___201505260615_MSG-N_______.h5"

    ct = MsgCloudType()
    ct.read(FILENAME)

    ct.convert2pps().save("blact.h5")

    FILENAME = "/data/proj/safutv/geo_out/0deg/SAFNWC_MSG3_CTTH_201505260615_MSG-N_______.h5"
    FILENAME = "/data/24/saf/geo_out/0deg/SAFNWC_MSG3_CTTH_201505260615_EuropeCanary.h5"
    ct = MsgCTTH()
    ct.read(FILENAME)

    ct.convert2pps().save("blactth.h5")



#!/usr/bin/env python

import urllib2
import h5py
import datetime as dt
import os.path
import sys

BASE_URL = "http://www.star.nesdis.noaa.gov/smcd/spb/fwu/homepage/" + \
           "AVHRR/Op_Cal_AVHRR/"

URLS = {
    "Metop-B":
    {"ch1": BASE_URL + "Metop1_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "Metop1_AVHRR_Libya_ch2.txt",
     "ch3a": BASE_URL + "Metop1_AVHRR_Libya_ch3a.txt"},
    "Metop-A":
    {"ch1": BASE_URL + "Metop2_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "Metop2_AVHRR_Libya_ch2.txt",
     "ch3a": BASE_URL + "Metop2_AVHRR_Libya_ch3a.txt"},
    "NOAA-16":
    {"ch1": BASE_URL + "N16_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N16_AVHRR_Libya_ch2.txt"},
    "NOAA-17":
    {"ch1": BASE_URL + "N17_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N17_AVHRR_Libya_ch2.txt",
     "ch3a": BASE_URL + "N17_AVHRR_Libya_ch3a.txt"},
    "NOAA-18":
    {"ch1": BASE_URL + "N18_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N18_AVHRR_Libya_ch2.txt"},
    "NOAA-19":
    {"ch1": BASE_URL + "N19_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N19_AVHRR_Libya_ch2.txt"}
}

def get_page(url):
    '''Retrieve the given page.'''
    return urllib2.urlopen(url).read()

def get_coeffs(page):
    '''Parse coefficients from the page.'''
    coeffs = {}
    coeffs['datetime'] = []
    coeffs['slope1'] = []
    coeffs['intercept1'] = []
    coeffs['slope2'] = []
    coeffs['intercept2'] = []

    slope1_idx, intercept1_idx, slope2_idx, intercept2_idx = \
        None, None, None, None

    date_idx = 0
    for row in page.lower().split('\n'):
        row = row.split()
        if len(row) == 0:
            continue
        if row[0] == 'update':
            # Get the column indices from the header line
            slope1_idx = row.index('slope_lo')
            intercept1_idx = row.index('int_lo')
            slope2_idx = row.index('slope_hi')
            intercept2_idx = row.index('int_hi')
            continue

        if slope1_idx is None:
            continue

        # In some cases the fields are connected, skip those rows
        if max([slope1_idx, intercept1_idx,
                slope2_idx, intercept2_idx]) >= len(row):
            continue

        try:
            dat = dt.datetime.strptime(row[date_idx], "%m/%d/%Y")
        except ValueError:
            continue

        coeffs['datetime'].append([dat.year, dat.month, dat.day])
        coeffs['slope1'].append(float(row[slope1_idx]))
        coeffs['intercept1'].append(float(row[intercept1_idx]))
        coeffs['slope2'].append(float(row[slope2_idx]))
        coeffs['intercept2'].append(float(row[intercept2_idx]))

    return coeffs

def get_all_coeffs():
    '''Get all available calibration coefficients for the satellites.'''
    coeffs = {}

    for platform in URLS.keys():
        if platform not in coeffs:
            coeffs[platform] = {}
        for chan in URLS[platform].keys():
            url = URLS[platform][chan]
            print url
            page = get_page(url)
            coeffs[platform][chan] = get_coeffs(page)

    return coeffs

def save_coeffs(coeffs, out_dir=''):
    '''Save calibration coefficients to HDF5 files.'''
    for platform in coeffs.keys():
        fname = os.path.join(out_dir, "%s_calibration_data.h5" % platform)
        fid = h5py.File(fname, 'w')
        
        for chan in coeffs[platform].keys():
            fid.create_group(chan)
            fid[chan]['datetime'] = coeffs[platform][chan]['datetime']
            fid[chan]['slope1'] = coeffs[platform][chan]['slope1']
            fid[chan]['intercept1'] = coeffs[platform][chan]['intercept1']
            fid[chan]['slope2'] = coeffs[platform][chan]['slope2']
            fid[chan]['intercept2'] = coeffs[platform][chan]['intercept2']

        fid.close()
        print "Calibration coefficients saved for %s" % platform

def main():
    '''Create calibration coefficient files for AVHRR'''
    out_dir = sys.argv[1]
    coeffs = get_all_coeffs()
    save_coeffs(coeffs, out_dir=out_dir)

if __name__ == "__main__":
    main()
